# Copyright 2024 Max Planck Institute for Software Systems, and
# National University of Singapore
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import annotations

from simbricks.orchestration.system import base
from simbricks.orchestration.system import pcie
from simbricks.orchestration.system import eth
from simbricks.utils import base as utils_base


class SimplePCIeNIC(base.Component):
    def __init__(self, s: base.System) -> None:
        super().__init__(s)
        self._ip: str | None = None
        self._eth_if: eth.EthInterface = eth.EthInterface(self)
        super().add_if(self._eth_if)
        self._pci_if: pcie.PCIeDeviceInterface = pcie.PCIeDeviceInterface(self)
        super().add_if(self._pci_if)

    def add_if(self, interface: eth.EthInterface | pcie.PCIeDeviceInterface) -> None:
        match interface:
            case eth.EthInterface():
                if hasattr(self, "_eth_if") and self._eth_if:
                    print(interface)
                    print(self._eth_if)
                    raise Exception(
                        f"you overwrite SimplePCIeNIC._eth_if ({self._eth_if.id()} -> {interface.id()}) "
                    )
                self._eth_if = interface
            case pcie.PCIeDeviceInterface():
                if hasattr(self, "_pci_if") and self._pci_if:
                    raise Exception(
                        f"you overwrite SimplePCIeNIC._pci_if. ({self._pci_if.id()} -> {interface.id()})"
                    )
                self._pci_if = interface
            case _:
                raise Exception(
                    f"interface must have type EthInterface or PCIeDeviceInterface but has type {type(interface)}"
                )
        super().add_if(interface)

    def add_ipv4(self, ip: str) -> None:
        assert self._ip is None
        self._ip = ip

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        json_obj["ip"] = self._ip
        json_obj["eth_if"] = self._eth_if.id()
        json_obj["pci_if"] = self._pci_if.id()
        return json_obj

    @classmethod
    def fromJSON(cls, system: base.System, json_obj: dict) -> SimplePCIeNIC:
        instance = super().fromJSON(system, json_obj)
        instance._ip = utils_base.get_json_attr_top(json_obj, "ip")
        eth_inf_id = int(utils_base.get_json_attr_top(json_obj, "eth_if"))
        instance._eth_if = system.get_inf(eth_inf_id)
        inf_id = int(utils_base.get_json_attr_top(json_obj, "pci_if"))
        instance._pci_if = system.get_inf(inf_id)
        return instance


class IntelI40eNIC(SimplePCIeNIC):
    def __init__(self, s: base.System) -> None:
        super().__init__(s)


class IntelE1000NIC(SimplePCIeNIC):
    def __init__(self, s: base.System) -> None:
        super().__init__(s)


class CorundumNIC(SimplePCIeNIC):
    def __init__(self, s: base.System) -> None:
        super().__init__(s)



# Copyright 2023 Max Planck Institute for Software Systems, and
# National University of Singapore
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

# Allow own class to be used as type for a method's argument
from __future__ import annotations

import typing as tp
from enum import Enum

import simbricks.orchestration.e2e_components as e2e
from simbricks.orchestration.simulators import NS3E2ENet


class E2ELinkType(Enum):
    SIMBRICKS = 0
    NS3_SIMPLE_CHANNEL = 1


class E2ELinkAssigner():

    def __init__(self):
        self.links = {}
        self.connected_switches = set()
        self.switch_links = {}

    def add_link(
        self,
        idd: str,
        left_switch: e2e.E2ETopologyNode,
        right_switch: e2e.E2ETopologyNode,
        link_type: tp.Optional[E2ELinkType] = None,
        create_link: bool = True
    ):
        if create_link and link_type is None:
            raise RuntimeError('Cannot create a link without link type')
        if idd in self.links:
            raise RuntimeError(f'Link {idd} already exists')
        link = {
            'left': left_switch,
            'right': right_switch,
            'type': link_type,
            'created': create_link
        }
        if create_link:
            self._create_link(idd, link)
        self.links[idd] = link
        self.connected_switches.add(left_switch)
        self.connected_switches.add(right_switch)
        if left_switch in self.switch_links:
            self.switch_links[left_switch].append(link)
        else:
            self.switch_links[left_switch] = [link]
        if right_switch in self.switch_links:
            self.switch_links[right_switch].append(link)
        else:
            self.switch_links[right_switch] = [link]

    # TODO: set properties like latency
    def _create_link(self, idd: str, link):
        left_switch = link['left']
        right_switch = link['right']
        link_type = link['type']
        if link_type == E2ELinkType.SIMBRICKS:
            left_adapter = e2e.E2ENetworkSimbricks(f'_{idd}_left_adapter')
            left_adapter.listen = False
            left_switch.add_component(left_adapter)
            link['left_adapter'] = left_adapter
            right_adapter = e2e.E2ENetworkSimbricks(f'_{idd}_right_adapter')
            right_adapter.listen = True
            right_switch.add_component(right_adapter)
            link['right_adapter'] = right_adapter
        elif link_type == E2ELinkType.NS3_SIMPLE_CHANNEL:
            ns3link = e2e.E2ESimpleChannel(f'_{idd}_link')
            ns3link.left_node = left_switch
            ns3link.right_node = right_switch
            link['ns3link'] = ns3link

    def set_link_type(self, idd: str, link_type: E2ELinkType):
        if idd not in self.links:
            raise RuntimeError(f'Link {idd} not found')
        link = self.links[idd]
        if link['created']:
            raise RuntimeError('Cannot change type of already existing link')
        link['type'] = link_type

    def create_missing_links(self):
        for idd, link in self.links.items():
            if link['created']:
                continue
            if link['type'] is None:
                raise RuntimeError(f'Link {idd} has no type')
            self._create_link(idd, link)
            link['created'] = True

    def assign_networks(self) -> tp.List[NS3E2ENet]:
        networks = []
        # walk over all connected switches
        while len(self.connected_switches) > 0:
            # create network and take next (random) switch
            net = NS3E2ENet()
            net.name = f'_network_{len(networks)}'
            networks.append(net)
            next_switches = set()
            next_switches.add(self.connected_switches.pop())
            # visit (transitively) all switches that are connected through ns3
            # links and add them to the same network
            while len(next_switches) > 0:
                switch = next_switches.pop()
                net.add_component(switch)
                for link in self.switch_links[switch]:
                    if link['type'] == E2ELinkType.SIMBRICKS:
                        if link['left'] == switch:
                            link['right_adapter'].simbricks_component = net
                        else:
                            assert link['right'] == switch
                            link['left_adapter'].simbricks_component = net
                    elif link['type'] == E2ELinkType.NS3_SIMPLE_CHANNEL:
                        not_visited_switches = 0
                        if link['left'] in self.connected_switches:
                            next_switches.add(link['left'])
                            self.connected_switches.remove(link['left'])
                            not_visited_switches += 1
                        if link['right'] in self.connected_switches:
                            next_switches.add(link['right'])
                            self.connected_switches.remove(link['right'])
                            not_visited_switches += 1
                        assert not_visited_switches < 2
                        # if only one switch has been visited (namely the
                        # current switch), we see this link for the first time
                        if not_visited_switches == 1:
                            net.add_component(link['ns3link'])
                    else:
                        raise RuntimeError('Unknown link type')

        return networks



# Copyright 2024 Max Planck Institute for Software Systems, and
# National University of Singapore
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import annotations

from simbricks.orchestration.system import base
from simbricks.utils import base as utils_base


class MemHostInterface(base.Interface):
    # Note AK: Component here is on purpose. Other simulators than host
    # simulators can also have MemHost interfaces (e.g. a Mem switch)
    def __init__(self, c: base.Component) -> None:
        super().__init__(c)


class MemDeviceInterface(base.Interface):
    def __init__(self, c: base.Component) -> None:
        super().__init__(c)

    def connect(self, c: base.Channel) -> None:
        # Note AK: a bit ugly, but I think we can't get around a rt check here
        if not c is isinstance(c, MemChannel):
            raise TypeError("MemDeviceInterface only connects to MemChannel")
        super().connect(c)


class MemChannel(base.Channel):
    def __init__(self, host: MemHostInterface, dev: MemDeviceInterface) -> None:
        super().__init__(host, dev)

    def host_if(self) -> MemHostInterface:
        return self.a

    def dev_if(self) -> MemDeviceInterface:
        return self.b


class MemSimpleDevice(base.Component):
    def __init__(self, s: base.System):
        super().__init__(s)
        self._mem_if: MemDeviceInterface = MemDeviceInterface(c=self)
        super().add_if(self._mem_if)
        self._addr = 0xE000000000000000
        self._size = 1024 * 1024 * 1024  # 1GB
        self._as_id = 0

    def add_if(self, interface: MemDeviceInterface) -> None:
        utils_base.has_expected_type(interface, MemDeviceInterface)
        if self._mem_if:
            raise Exception(
                f"you overwrite MemDeviceInterface._mem_if ({self._mem_if.id()} -> {interface.id()}) "
            )
        self._mem_if = interface
        super().add_if(self._mem_if)

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        json_obj["mem_if"] = self._mem_if.id()
        json_obj["addr"] = self._addr
        json_obj["size"] = self._size
        json_obj["as_id"] = self._as_id
        return json_obj

    @classmethod
    def fromJSON(cls, system: base.System, json_obj: dict) -> MemSimpleDevice:
        instance = super().fromJSON(system, json_obj)
        mem_if_id = int(utils_base.get_json_attr_top(json_obj, "mem_if"))
        addr = utils_base.get_json_attr_top(json_obj, "addr")
        size = utils_base.get_json_attr_top(json_obj, "size")
        as_id = utils_base.get_json_attr_top(json_obj, "as_id")
        instance.mem_if = system.get_inf(mem_if_id)
        instance.addr = addr
        instance.size = size
        instance.as_id = as_id
        return instance



# Copyright 2024 Max Planck Institute for Software Systems, and
# National University of Singapore
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import annotations

import typing as tp
import io
import asyncio
import simbricks.orchestration.instantiation.base as instantiation
from simbricks.orchestration.system import base as base
from simbricks.orchestration.system import eth as eth
from simbricks.orchestration.system import nic as nic
from simbricks.orchestration.system import pcie as pcie
from simbricks.orchestration.system.host import app
from simbricks.utils import base as utils_base

if tp.TYPE_CHECKING:
    from simbricks.orchestration.system.host import disk_images


class Host(base.Component):
    def __init__(self, s: base.System):
        super().__init__(s)
        self.applications: list[app.Application] = []

    def add_app(self, a: app.Application) -> None:
        self.applications.append(a)

    def toJSON(self) -> dict:
        json_obj = super().toJSON()

        applications_json = []
        for app in self.applications:
            utils_base.has_attribute(app, "toJSON")
            applications_json.append(app.toJSON())
        json_obj["applications"] = applications_json

        return json_obj

    @classmethod
    def fromJSON(cls, system: base.System, json_obj: dict) -> Host:
        instance = super().fromJSON(system=system, json_obj=json_obj)
        instance.applications = []

        applications_json = utils_base.get_json_attr_top(json_obj, "applications")
        for app_json in applications_json:
            app_class = utils_base.get_cls_by_json(app_json)
            utils_base.has_attribute(app_class, "fromJSON")
            app = app_class.fromJSON(system, app_json)
            instance.add_app(app)

        return instance

    def connect_pcie_dev(self, dev: pcie.PCIeSimpleDevice) -> pcie.PCIeChannel:
        pcie_if = pcie.PCIeHostInterface(self)
        self.add_if(pcie_if)
        pcichannel0 = pcie.PCIeChannel(pcie_if, dev._pci_if)
        return pcichannel0


class FullSystemHost(Host):
    def __init__(self, s: base.System) -> None:
        super().__init__(s)
        self.memory: int = 2048
        self.cores: int = 1
        self.cpu_freq: str = "3GHz"
        self.disks: list[disk_images.DiskImage] = []

    def add_disk(self, disk: disk_images.DiskImage) -> None:
        self.disks.append(disk)

    async def prepare(self, inst: instantiation.Instantiation) -> None:
        promises = [disk.prepare(inst) for disk in self.disks]
        await asyncio.gather(*promises)

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        json_obj["memory"] = self.memory
        json_obj["cores"] = self.cores
        json_obj["cpu_freq"] = self.cpu_freq

        disks_json = []
        for disk in self.disks:
            utils_base.has_attribute(disk, "toJSON")
            disks_json.append(disk.toJSON())
        json_obj["disks"] = disks_json

        return json_obj

    @classmethod
    def fromJSON(cls, system: base.System, json_obj: dict) -> FullSystemHost:
        instance = super().fromJSON(system, json_obj)
        instance.memory = int(utils_base.get_json_attr_top(json_obj, "memory"))
        instance.cores = int(utils_base.get_json_attr_top(json_obj, "cores"))
        instance.cpu_freq = utils_base.get_json_attr_top(json_obj, "cpu_freq")

        instance.disks = []
        disks_json = utils_base.get_json_attr_top(json_obj, "disks")
        for disk_js in disks_json:
            disk_class = utils_base.get_cls_by_json(disk_js)
            utils_base.has_attribute(disk_class, "fromJSON")
            disk = disk_class.fromJSON(system, disk_js)
            instance.add_disk(disk)

        return instance


class BaseLinuxHost(FullSystemHost):
    def __init__(self, s: base.System) -> None:
        super().__init__(s)
        self.applications: list[app.BaseLinuxApplication] = []
        self.load_modules = []
        self.kcmd_append: str | None = None

    def add_app(self, a: app.BaseLinuxApplication) -> None:
        self.applications.append(a)

    def _concat_app_cmds(
        self,
        inst: instantiation.Instantiation,
        mapper_name: str,
    ) -> list[str]:
        """
        Generate command list from applications by applying `mapper` to each
        application on this host and concatenating the commands.
        """
        cmds = []
        for app in self.applications:
            mapper = getattr(app, mapper_name, None)
            if mapper is None:
                raise Exception(
                    f"coulkd not determine mapper function with name {mapper_name}"
                )
            cmds += mapper(inst)

        return cmds

    def run_cmds(self, inst: instantiation.Instantiation) -> list[str]:
        """Commands to run on node."""
        return self._concat_app_cmds(inst, app.BaseLinuxApplication.run_cmds.__name__)

    def cleanup_cmds(self, inst: instantiation.Instantiation) -> list[str]:
        """Commands to run to cleanup node."""
        cmds = self._concat_app_cmds(
            inst, app.BaseLinuxApplication.cleanup_cmds.__name__
        )
        sim = inst.find_sim_by_spec(spec=self)
        cleanup = sim.cleanup_commands()
        cmds += cleanup
        return cmds

    def config_files(self, inst: instantiation.Instantiation) -> dict[str, tp.IO]:
        """
        Additional files to put inside the node, which are mounted under
        `/tmp/guest/`.

        Specified in the following format: `filename_inside_node`:
        `IO_handle_of_file`
        """
        cfg_files = {}
        for app in self.applications:
            cfg_files |= app.config_files(inst)
        return cfg_files

    def prepare_pre_cp(self, inst: instantiation.Instantiation) -> list[str]:
        """Commands to run to prepare node before checkpointing."""
        return self._concat_app_cmds(
            inst, app.BaseLinuxApplication.prepare_pre_cp.__name__
        )

    def prepare_post_cp(self, inst: instantiation.Instantiation) -> list[str]:
        """Commands to run to prepare node after checkpoint restore."""
        return self._concat_app_cmds(
            inst, app.BaseLinuxApplication.prepare_post_cp.__name__
        )

    def config_str(self, inst: instantiation.Instantiation) -> str:
        sim = inst.find_sim_by_spec(spec=self)
        if inst.create_checkpoint:
            cp_cmd = sim.checkpoint_commands()
        else:
            cp_cmd = []

        es = (
            self.prepare_pre_cp(inst)
            + self.applications[0].prepare_pre_cp(inst)
            + cp_cmd
            + self.prepare_post_cp(inst)
            + self.applications[0].prepare_post_cp(inst)
            + self.run_cmds(inst)
            + self.cleanup_cmds(inst)
        )
        cmd = "\n".join(es)
        return cmd

    def strfile(self, s: str) -> io.BytesIO:
        """
        Helper function to convert a string to an IO handle for usage in
        `config_files()`.

        Using this, you can create a file with the string as its content on the
        simulated node.
        """
        return io.BytesIO(bytes(s, encoding="UTF-8"))

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        json_obj["load_modules"] = self.load_modules
        json_obj["kcmd_append"] = self.kcmd_append
        return json_obj

    @classmethod
    def fromJSON(cls, system: base.System, json_obj: dict) -> BaseLinuxHost:
        instance = super().fromJSON(system, json_obj)
        instance.load_modules = utils_base.get_json_attr_top(json_obj, "load_modules")
        instance.kcmd_append = utils_base.get_json_attr_top(json_obj, "kcmd_append")
        return instance


class LinuxHost(BaseLinuxHost):
    def __init__(self, sys) -> None:
        super().__init__(sys)
        self.drivers: list[str] = []
        self.hostname: str | None = "ubuntu"

    def cleanup_cmds(self, inst: instantiation.Instantiation) -> list[str]:
        return super().cleanup_cmds(inst) + ["poweroff -f"]

    def prepare_pre_cp(self, inst: instantiation.Instantiation) -> list[str]:
        """Commands to run to prepare node before checkpointing."""
        cmds = [
            "set -x",
            "export HOME=/root",
            "export LANG=en_US",
            'export PATH="/usr/local/sbin:/usr/local/bin:/usr/sbin:'
            + '/usr/bin:/sbin:/bin:/usr/games:/usr/local/games"',
        ]
        if self.hostname is not None:
            cmds += [
                f"hostname -b {self.hostname}",
                f'echo "127.0.1.1 {self.hostname}\n" >> /etc/hosts',
            ]
        cmds += super().prepare_pre_cp(inst)
        return cmds

    def prepare_post_cp(self, inst) -> list[str]:
        cmds = super().prepare_post_cp(inst)
        for d in self.drivers:
            if d[0] == "/":
                cmds.append(f"insmod {d}")
            else:
                cmds.append(f"modprobe {d}")

        index = 0
        for host_inf in base.Interface.filter_by_type(
            self.interfaces(), pcie.PCIeHostInterface
        ):
            if not host_inf.is_connected():
                continue

            inf = host_inf.get_opposing_interface()
            if not utils_base.check_types(
                inf.component, eth.EthSimpleNIC, nic.SimplePCIeNIC
            ):
                continue
            # Get ifname parameter if set, otherwise default to ethX
            ifn = f"eth{index}"
            index += 1
            com: eth.EthSimpleNIC | nic.SimplePCIeNIC = inf.component

            # Force MAC if requested TODO: FIXME
            # if "force_mac_addr" in i.parameters:
            #     mac = i.parameters["force_mac_addr"]
            #     l.append(f"ip link set dev {ifn} address " f"{mac}")

            # Bring interface up
            cmds.append(f"ip link set dev {ifn} up")

            # Add IP addresses if included
            assert com._ip is not None
            cmds.append(f"ip addr add {com._ip}/24 dev {ifn}")

        return cmds

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        json_obj["drivers"] = self.drivers
        json_obj["hostname"] = self.hostname
        return json_obj

    @classmethod
    def fromJSON(cls, system: base.System, json_obj: dict) -> LinuxHost:
        instance = super().fromJSON(system, json_obj)
        instance.drivers = utils_base.get_json_attr_top(json_obj, "drivers")
        instance.hostname = utils_base.get_json_attr_top(json_obj, "hostname")
        return instance


class I40ELinuxHost(LinuxHost):
    def __init__(self, sys) -> None:
        super().__init__(sys)
        self.drivers.append("i40e")


class CorundumLinuxHost(LinuxHost):
    def __init__(self, sys) -> None:
        super().__init__(sys)
        self.drivers.append("/tmp/guest/mqnic.ko")

    def config_files(self, inst: instantiation.Instantiation) -> tp.Dict[str, tp.IO]:
        m = {
            "mqnic.ko": open(
                f"{inst.join_repo_base(relative_path='images/mqnic/mqnic.ko')}", "rb"
            )
        }
        return {**m, **super().config_files(inst=inst)}



# Copyright 2023 Max Planck Institute for Software Systems, and
# National University of Singapore
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

# Allow own class to be used as type for a method's argument
from __future__ import annotations

import typing as tp
from abc import ABC, abstractmethod
from enum import Enum


class CongestionControl(Enum):
    RENO = ('ns3::TcpLinuxReno', 'reno')
    WESTWOOD = ('ns3::TcpWestwoodPlus', 'westwood')
    BIC = ('ns3::TcpBic', 'bic')
    CUBIC = ('ns3::TcpCubic', 'cubic')
    HTCP = ('ns3::TcpHtcp', 'htcp')
    HYBLA = ('ns3::TcpHybla', 'hybla')
    VEGAS = ('ns3::TcpVegas', 'vegas')
    NV = ('', 'nv')
    SCALABLE = ('ns3::TcpScalable', 'scalable')
    LP = ('ns3::TcpLp', 'lp')
    VENO = ('ns3::TcpVeno', 'veno')
    YEAH = ('ns3::TcpYeah', 'yeah')
    ILLINOIS = ('ns3::TcpIllinois', 'illinois')
    DCTCP = ('ns3::TcpDctcp', 'dctcp')
    CDG = ('', 'cdg')
    BBR = ('ns3::TcpBbr', 'bbr')
    HIGHSPEED = ('ns3::TcpHighSpeed', 'highspeed')

    def __init__(self, ns3_str, gem5_str):
        self.ns3_str = ns3_str
        self.gem5_str = gem5_str

    def __str__(self):
        return self.name.lower()

    @property
    def ns3(self):
        if self.ns3_str == '':
            raise AttributeError(
                f'There is no ns3 implementation for '
                f'{self.name} available'
            )
        return self.ns3_str

    @property
    def gem5(self):
        return self.gem5_str


class SimbricksAdapterType(Enum):
    NIC = 0
    NETWORK = 1
    HOST = 2


class SimbricksSyncMode(Enum):
    SYNC_DISABLED = 0
    SYNC_OPTIONAL = 1
    SYNC_REQUIRED = 2


class Ns3LoggingLevel(Enum):
    ERROR = 'error'
    LEVEL_ERROR = 'level_error'
    WARN = 'warn'
    LEVEL_WARN = 'level_warn'
    DEBUG = 'debug'
    LEVEL_DEBUG = 'level_debug'
    INFO = 'info'
    LEVEL_INFO = 'level_info'
    FUNCTION = 'function'
    LEVEL_FUNCTION = 'level_function'
    LOGIC = 'logic'
    LEVEL_LOGIC = 'level_logic'
    ALL = 'all'
    LEVEL_ALL = 'level_all'
    PREFIX_FUNC = 'prefix_func'
    PREFIX_TIME = 'prefix_time'
    PREFIX_NODE = 'prefix_node'
    PREFIX_LEVEL = 'prefix_level'
    PREFIX_ALL = 'prefix_all'


class E2EBase(ABC):

    def __init__(self) -> None:
        self.category: str
        self.mapping: tp.Dict[str, str] = {}
        self.components: tp.List[E2EComponent] = []

    def ns3_config(self) -> str:
        config_list = []
        for key, value in self.mapping.items():
            if value == '':
                continue
            config_list.append(f'{key}:{value}')
        config = ';'.join(config_list)

        child_configs = ' '.join([
            child.ns3_config() for child in self.components
        ])

        return f'--{self.category}="{config}" {child_configs}'

    @abstractmethod
    def add_component(self, component: E2EComponent) -> None:
        pass


class E2EGlobalConfig(E2EBase):

    def __init__(self) -> None:
        super().__init__()
        self.category = 'Global'
        self.stop_time = ''
        self.mac_start = 0

    def ns3_config(self) -> str:
        self.mapping.update({
            'StopTime': self.stop_time, 'MACStart': str(self.mac_start)
        })
        return super().ns3_config()

    def add_component(self, component: E2EComponent) -> None:
        raise AttributeError("Can't add a component to the global config")


class E2ENs3Logging(E2EBase):

    def __init__(self) -> None:
        super().__init__()
        self.category = 'Logging'
        self.logging: tp.Dict[str, tp.List[Ns3LoggingLevel]] = {}

    def ns3_config(self) -> str:
        for component, levels in self.logging.items():
            levels_str = '|'.join([level.value for level in levels])
            self.mapping.update([(component, levels_str)])
        return super().ns3_config()

    def add_component(self, component: E2EComponent) -> None:
        raise AttributeError("Can't add a component to the global config")

    def add_logging(self, component: str, level: Ns3LoggingLevel):
        if component in self.logging:
            self.logging[component].append(level)
        else:
            self.logging[component] = [level]


class E2EComponent(E2EBase):

    def __init__(self, idd: str) -> None:
        super().__init__()
        self.name = idd
        self.id = idd
        self.has_path = False
        self.type = ''

    def ns3_config(self) -> str:
        if self.id == '' or self.type == '':
            raise AttributeError('Id or Type cannot be empty')
        self.mapping.update({'Id': self.id, 'Type': self.type})

        return super().ns3_config()

    def add_component(self, component: E2EComponent) -> None:
        self.components.append(component)

    def resolve_paths(self) -> None:
        self.has_path = True
        for component in self.components:
            path = f'{self.id}/{component.id}'
            if component.has_path:
                raise AttributeError(
                    f'Component {component.id} was already '
                    f'added to another component (while trying '
                    f'to assign {path}).'
                )
            component.id = path
            component.resolve_paths()


class E2ETopologyNode(E2EComponent):

    def __init__(self, idd: str) -> None:
        super().__init__(idd)
        self.category = 'TopologyNode'
        self.network = None


class E2ESwitchNode(E2ETopologyNode):

    def __init__(self, idd: str) -> None:
        super().__init__(idd)
        self.type = 'Switch'
        self.mtu = ''

    def ns3_config(self) -> str:
        self.mapping.update({
            'Mtu': self.mtu,
        })
        return super().ns3_config()


class E2ETopologyChannel(E2EComponent):

    def __init__(self, idd: str) -> None:
        super().__init__(idd)
        self.category = 'TopologyChannel'


class E2ESimpleChannel(E2ETopologyChannel):

    def __init__(self, idd: str) -> None:
        super().__init__(idd)
        self.type = 'Simple'
        self.data_rate = ''
        self.queue_type = 'ns3::DropTailQueue'
        self.queue_size = ''
        self.channel_type = 'ns3::SimpleChannel'
        self.delay = ''
        self.left_node: E2ETopologyNode
        self.right_node: E2ETopologyNode

    def ns3_config(self) -> str:
        if self.left_node is None or self.right_node is None:
            raise AttributeError(f'Not all nodes for channel {self.id} given')
        self.mapping.update({
            'Device-DataRate': self.data_rate,
            'QueueType': self.queue_type,
            'Queue-MaxSize': self.queue_size,
            'ChannelType': self.channel_type,
            'Channel-Delay': self.delay,
            'LeftNode': self.left_node.id,
            'RightNode': self.right_node.id,
        })
        return super().ns3_config()

    def add_device_attr(self, key: str, value: str) -> None:
        if not key.startswith('Device-'):
            key = f'Device-{key}'
        self.mapping.update({key: value})

    def add_queue_attr(self, key: str, value: str) -> None:
        if not key.startswith('Queue-'):
            key = f'Queue-{key}'
        self.mapping.update({key: value})

    def add_channel_attr(self, key: str, value: str) -> None:
        if not key.startswith('Channel-'):
            key = f'Channel-{key}'
        self.mapping.update({key: value})


class E2ENetwork(E2EComponent):

    def __init__(self, idd: str) -> None:
        super().__init__(idd)
        self.category = 'Network'
        self.peer = None

    def set_peer(self, peer: E2ENetwork):
        self.peer = peer
        peer.peer = self


class E2ENetworkSimbricks(E2ENetwork):

    def __init__(self, idd: str) -> None:
        super().__init__(idd)
        self.type = 'Simbricks'
        self.adapter_type = SimbricksAdapterType.NETWORK
        self.unix_socket = ''
        self.sync_delay = ''
        self.poll_delay = ''
        self.eth_latency = ''
        self.listen: tp.Optional[bool] = None
        self.shm_path = ''
        self.sync: SimbricksSyncMode = SimbricksSyncMode.SYNC_OPTIONAL

        self.simbricks_component = None

    def ns3_config(self) -> str:
        if self.listen is None:
            raise AttributeError(
                f'Listen mode not specified for simbricks adapter {self.id}'
            )
        self.mapping.update({
            'UnixSocket': self.unix_socket,
            'SyncDelay': self.sync_delay,
            'PollDelay': self.poll_delay,
            'EthLatency': self.eth_latency,
            'Listen': 'true' if self.listen else 'false',
            'ShmPath': self.shm_path,
            'Sync': '' if self.sync is None else f'{self.sync.value}',
        })
        return super().ns3_config()


class E2EHost(E2EComponent):

    def __init__(self, idd: str) -> None:
        super().__init__(idd)
        self.category = 'Host'


class E2ESimbricksHost(E2EHost):

    def __init__(self, idd: str) -> None:
        super().__init__(idd)
        self.type = 'Simbricks'
        self.adapter_type = SimbricksAdapterType.NIC
        self.unix_socket = ''
        self.sync_delay = ''
        self.poll_delay = ''
        self.eth_latency = ''
        self.sync: SimbricksSyncMode = SimbricksSyncMode.SYNC_OPTIONAL

        self.simbricks_component = None

    def ns3_config(self) -> str:
        self.mapping.update({
            'UnixSocket': self.unix_socket,
            'SyncDelay': self.sync_delay,
            'PollDelay': self.poll_delay,
            'EthLatency': self.eth_latency,
            'Sync': '' if self.sync is None else f'{self.sync.value}',
        })
        return super().ns3_config()


class E2ESimpleNs3Host(E2EHost):

    def __init__(self, idd: str) -> None:
        super().__init__(idd)
        self.type = 'SimpleNs3'
        self.data_rate = ''
        self.queue_type = 'ns3::DropTailQueue'
        self.queue_size = ''
        self.channel_type = 'ns3::SimpleChannel'
        self.delay = ''
        self.congestion_control: CongestionControl = None
        self.ip = ''

    def ns3_config(self) -> str:
        if self.congestion_control is None:
            cc = ''
        else:
            cc = self.congestion_control.ns3
        self.mapping.update({
            'Device-DataRate': self.data_rate,
            'QueueType': self.queue_type,
            'Queue-MaxSize': self.queue_size,
            'ChannelType': self.channel_type,
            'Channel-Delay': self.delay,
            'CongestionControl': cc,
            'Ip': self.ip,
        })
        return super().ns3_config()

    def add_device_attr(self, key: str, value: str) -> None:
        if not key.startswith('Device-'):
            key = f'Device-{key}'
        self.mapping.update({key: value})

    def add_queue_attr(self, key: str, value: str) -> None:
        if not key.startswith('Queue-'):
            key = f'Queue-{key}'
        self.mapping.update({key: value})

    def add_channel_attr(self, key: str, value: str) -> None:
        if not key.startswith('Channel-'):
            key = f'Channel-{key}'
        self.mapping.update({key: value})


class E2EApplication(E2EComponent):

    def __init__(self, idd: str) -> None:
        super().__init__(idd)
        self.category = 'App'
        self.start_time = ''
        self.stop_time = ''

    def ns3_config(self) -> str:
        self.mapping.update({
            'StartTime': self.start_time,
            'StopTime': self.stop_time,
        })
        return super().ns3_config()


class E2EPacketSinkApplication(E2EApplication):

    def __init__(self, idd: str) -> None:
        super().__init__(idd)
        self.type = 'PacketSink'
        self.protocol = 'ns3::TcpSocketFactory'
        self.local_ip = ''

    def ns3_config(self) -> str:
        self.mapping.update({
            'Protocol': self.protocol,
            'Local': self.local_ip,
        })
        return super().ns3_config()


class E2EBulkSendApplication(E2EApplication):

    def __init__(self, idd: str) -> None:
        super().__init__(idd)
        self.type = 'BulkSender'
        self.protocol = 'ns3::TcpSocketFactory'
        self.remote_ip = ''

    def ns3_config(self) -> str:
        self.mapping.update({
            'Protocol': self.protocol,
            'Remote': self.remote_ip,
        })
        return super().ns3_config()


class E2ENs3RandomVariable(ABC):

    def __init__(self) -> None:
        self.type_id = ''

    def get_config(self) -> str:
        params = self.get_parameters()
        if params:
            return f'{self.type_id}[{params}]'
        else:
            return self.type_id

    @abstractmethod
    def get_parameters(self) -> str:
        pass


class E2ENs3ConstantRandomVariable(E2ENs3RandomVariable):

    def __init__(self) -> None:
        super().__init__()
        self.type_id = 'ns3::ConstantRandomVariable'
        self.constant: tp.Optional[float] = None

    def get_parameters(self) -> str:
        params = []
        if self.constant:
            params.append(f'Constant={self.constant}')
        return '|'.join(params)


class E2ENs3UniformRandomVariable(E2ENs3RandomVariable):

    def __init__(self) -> None:
        super().__init__()
        self.type_id = 'ns3::UniformRandomVariable'
        self.min: tp.Optional[float] = None
        self.max: tp.Optional[float] = None

    def get_parameters(self) -> str:
        params = []
        if self.min:
            params.append(f'Min={self.min}')
        if self.max:
            params.append(f'Max={self.max}')
        return '|'.join(params)


class E2ENs3ExponentialRandomVariable(E2ENs3RandomVariable):

    def __init__(self) -> None:
        super().__init__()
        self.type_id = 'ns3::ExponentialRandomVariable'
        self.mean: tp.Optional[float] = None
        self.bound: tp.Optional[float] = None

    def get_parameters(self) -> str:
        params = []
        if self.mean:
            params.append(f'Mean={self.mean}')
        if self.bound:
            params.append(f'Bound={self.bound}')
        return '|'.join(params)


class E2ENs3NormalRandomVariable(E2ENs3RandomVariable):

    def __init__(self) -> None:
        super().__init__()
        self.type_id = 'ns3::NormalRandomVariable'
        self.mean: tp.Optional[float] = None
        self.variance: tp.Optional[float] = None
        self.bound: tp.Optional[float] = None

    def get_parameters(self) -> str:
        params = []
        if self.mean:
            params.append(f'Mean={self.mean}')
        if self.variance:
            params.append(f'Variance={self.variance}')
        if self.bound:
            params.append(f'Bound={self.bound}')
        return '|'.join(params)


class E2EOnOffApplication(E2EApplication):

    def __init__(self, idd: str) -> None:
        super().__init__(idd)
        self.type = 'OnOff'
        self.protocol = 'ns3::TcpSocketFactory'
        self.remote_ip = ''
        self.data_rate = ''
        self.max_bytes = ''
        self.packet_size = ''
        self.on_time: tp.Optional[E2ENs3RandomVariable] = None
        self.off_time: tp.Optional[E2ENs3RandomVariable] = None

    def ns3_config(self) -> str:
        if self.on_time:
            on = self.on_time.get_config()
        else:
            on = ''
        if self.off_time:
            off = self.off_time.get_config()
        else:
            off = ''
        self.mapping.update({
            'Protocol': self.protocol,
            'Remote': self.remote_ip,
            'DataRate': self.data_rate,
            'MaxBytes': self.max_bytes,
            'PacketSize': self.packet_size,
            'OnTime': on,
            'OffTime': off,
        })
        return super().ns3_config()


class E2EProbe(E2EComponent):

    def __init__(self, idd: str) -> None:
        super().__init__(idd)
        self.category = 'Probe'


class E2EPeriodicSampleProbe(E2EProbe):

    def __init__(self, idd: str, probe_type: str) -> None:
        super().__init__(idd)
        self.type = probe_type
        self.file = ''
        self.header = ''
        self.unit = ''
        self.start = ''
        self.interval = ''

    def ns3_config(self) -> str:
        self.mapping.update({
            'File': self.file,
            'Header': self.header,
            'Unit': self.unit,
            'Start': self.start,
            'Interval': self.interval
        })
        return super().ns3_config()



# Copyright 2024 Max Planck Institute for Software Systems, and
# National University of Singapore
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from simbricks.orchestration import system
from simbricks.orchestration.simulation import base as sim_base
from simbricks.orchestration.simulation import channel as sim_chan
from simbricks.orchestration.instantiation import base as inst_base
from simbricks.utils import base as utils_base

def simple_instantiation(
    simulation: sim_base.Simulation,
) -> inst_base.Instantiation:
  instance = inst_base.Instantiation(sim=simulation)
  instance.preserve_tmp_folder = False
  instance.create_checkpoint = True
  return instance


# Copyright 2024 Max Planck Institute for Software Systems, and
# National University of Singapore
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

import typing
import functools
from simbricks.utils import base as util_base
from simbricks.orchestration.instantiation import proxy
from simbricks.utils import base as util_base

if typing.TYPE_CHECKING:
    from simbricks.orchestration.simulation import base as sim_base
    from simbricks.orchestration.system import base as sys_base


class Fragment(util_base.IdObj):

    def __init__(self):
        super().__init__()

        self._proxies: set[proxy.Proxy] = set()
        self._simulators: set[sim_base.Simulator] = set()

    def toJSON(self) -> dict:
        json_obj = super().toJSON()

        proxy_json = []
        for prox in self._proxies:
            util_base.has_attribute(prox, "toJSON")
            proxy_json.append(prox.toJSON())
        json_obj["proxies"] = proxy_json

        json_obj["simulators"] = list(map(lambda sim: sim.id(), self._simulators))
        json_obj["cores_required"] = self.cores_required
        json_obj["memory_required"] = self.memory_required
        return json_obj

    @property
    def cores_required(self) -> int:
        req_cores_per_sim = map(lambda sim: sim.resreq_cores(), self._simulators)
        req_cores = functools.reduce(lambda x, y: x + y, req_cores_per_sim)
        return req_cores

    @property
    def memory_required(self) -> int:
        req_mem_per_sim = map(lambda sim: sim.resreq_mem(), self._simulators)
        req_mem = functools.reduce(lambda x, y: x + y, req_mem_per_sim)
        return req_mem

    @classmethod
    def fromJSON(cls, json_obj: dict) -> Fragment:
        instance = super().fromJSON(json_obj)
        # TODO: FIXME implement proper reconstruction from json
        return instance

    @staticmethod
    def merged(*fragments: Fragment):
        merged_fragment = Fragment()
        proxies = set()
        simulators = set()
        for fragment in fragments:
            proxies.update(fragment.all_proxies())
            simulators.update(fragment.all_simulators())
        merged_fragment._proxies = proxies
        merged_fragment._simulators = simulators

    def add_simulators(self, *sims: sim_base.Simulator):
        self._simulators.update(sims)

    def add_proxies(self, *proxies: proxy.Proxy):
        self._proxies.update(proxies)

    def all_simulators(self) -> set[sim_base.Simulator]:
        return self._simulators

    def all_proxies(self) -> set[proxy.Proxy]:
        return self._proxies

    def find_proxy_by_interface(self, interface: sys_base.Interface) -> proxy.Proxy | None:
        for proxy in self._proxies:
            if interface in proxy.interfaces:
                return proxy
        return None

    def get_proxy_by_interface(self, interface: sys_base.Interface) -> proxy.Proxy:
        """Same as `find_proxy_by_interface()` but raises an Error if interface
        is assigned to any proxy in this fragment."""
        proxy = self.find_proxy_by_interface(interface)
        if proxy is None:
            raise RuntimeError("Interface not assigned to any proxies in this fragment.")
        return proxy

    def interface_handled_by_proxy(self, interface: sys_base.Interface) -> bool:
        return self.find_proxy_by_interface(interface) is not None



# Copyright 2023 Max Planck Institute for Software Systems, and
# National University of Singapore
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

# Allow own class to be used as type for a method's argument
from __future__ import annotations

from abc import ABC, abstractmethod

import simbricks.orchestration.e2e_components as e2e


class E2ETopology(ABC):

    @abstractmethod
    def add_to_network(self, net):
        pass


class E2EDumbbellTopology(E2ETopology):

    def __init__(self):
        self.left_switch = e2e.E2ESwitchNode('_leftSwitch')
        self.right_switch = e2e.E2ESwitchNode('_rightSwitch')
        self.link = e2e.E2ESimpleChannel('_link')
        self.link.left_node = self.left_switch
        self.link.right_node = self.right_switch

    def add_to_network(self, net):
        net.add_component(self.left_switch)
        net.add_component(self.right_switch)
        net.add_component(self.link)

    def add_left_component(self, component: e2e.E2EComponent):
        self.left_switch.add_component(component)

    def add_right_component(self, component: e2e.E2EComponent):
        self.right_switch.add_component(component)

    @property
    def mtu(self):
        return self.left_switch.mtu

    @mtu.setter
    def mtu(self, mtu: str):
        self.left_switch.mtu = mtu
        self.right_switch.mtu = mtu

    @property
    def data_rate(self):
        return self.link.data_rate

    @data_rate.setter
    def data_rate(self, data_rate: str):
        self.link.data_rate = data_rate

    @property
    def queue_size(self):
        return self.link.queue_size

    @queue_size.setter
    def queue_size(self, queue_size: str):
        self.link.queue_size = queue_size

    @property
    def delay(self):
        return self.link.delay

    @delay.setter
    def delay(self, delay: str):
        self.link.delay = delay



# Copyright 2024 Max Planck Institute for Software Systems, and
# National University of Singapore
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import annotations


from simbricks.orchestration.simulation import base as sim_base
from simbricks.orchestration.system import base as system_base
from simbricks.utils import base as utils_base


class Channel(utils_base.IdObj):

    def __init__(self, chan: system_base.Channel):
        super().__init__()
        self._synchronized: bool = False
        self.sync_period: int = 500  # nanoseconds
        """
        The synchronization period in nanoseconds. For SimBricks to function 
        properly in sync mode, the sync period must not be larger than a channels 
        latency.
        """
        assert self.sync_period <= chan.latency
        self.sys_channel: system_base.Channel = chan

    def toJSON(self):
        json_obj = super().toJSON()
        json_obj["synchronized"] = self._synchronized
        json_obj["sync_period"] = self.sync_period
        json_obj["sys_channel"] = self.sys_channel.id()
        return json_obj

    @classmethod
    def fromJSON(cls, simulation: sim_base.Simulation, json_obj: dict) -> Channel:
        instance = super().fromJSON(json_obj)
        instance._synchronized = bool(utils_base.get_json_attr_top(json_obj, "synchronized"))
        instance.sync_period = int(utils_base.get_json_attr_top(json_obj, "sync_period"))
        chan_id = int(utils_base.get_json_attr_top(json_obj, "sys_channel"))
        instance.sys_channel = simulation.system.get_chan(chan_id)
        return instance

    def full_name(self) -> str:
        return "channel." + self.name

    def set_sync_period(
        self, amount: int, ratio: utils_base.Time = utils_base.Time.Nanoseconds
    ) -> None:
        utils_base.has_expected_type(obj=ratio, expected_type=utils_base.Time)
        self.sync_period = amount * ratio
        assert self.sync_period <= self.sys_channel.latency



# Copyright 2022 Max Planck Institute for Software Systems, and
# National University of Singapore
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import annotations

import typing

import simbricks.utils.base as util_base
from simbricks.orchestration.instantiation import socket as inst_socket
import abc
import typing

if typing.TYPE_CHECKING:
    from simbricks.orchestration.instantiation import base as inst_base
    import simbricks.orchestration.system.base as sys_base


class Proxy(util_base.IdObj, abc.ABC):

    def __init__(self):
        super().__init__()
        self._interfaces: list[sys_base.Interface]
        """
        The interfaces this proxy provides.
        
        Order is important here because proxies forward messages for SimBricks
        sockets in the order these sockets are passed on the command-line. So
        for two connecting proxies executing on separate runners, this order
        must be the same.
        """
        self.connection_mode: inst_socket.SockType = inst_socket.SockType.CONNECT

    @property
    def name(self) -> str:
        return f"proxy_{self.id()}"

    @abc.abstractmethod
    def run_cmd() -> str:
        pass

    def sockets_wait(self, inst: inst_base.Instantiation) -> set[inst_socket.Socket]:
        wait_sockets = []
        for iface in self._interfaces:
            socket = inst.update_get_socket(iface)
            if socket.type == inst_socket.SockType.LISTEN:
                wait_sockets.append(socket)
        return wait_sockets

    def supported_socket_types(
        interface: sys_base.Interface,
    ) -> set[inst_socket.SockType]:
        return {inst_socket.SockType.CONNECT, inst_socket.SockType.LISTEN}


class DummyProxy(Proxy):

    def __init__(self):
        super().__init__()


class TCPProxy(Proxy):

    def __init__(self):
        super().__init__()
        self.ip: str
        self.port: int


class RDMAProxy(Proxy):

    def __init__(self):
        super().__init__()
        self.ip: str
        self.port: int



# Copyright 2024 Max Planck Institute for Software Systems, and
# National University of Singapore
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import annotations

import math
import simbricks.orchestration.simulation.base as sim_base
import simbricks.orchestration.system as system
from simbricks.orchestration.instantiation import base as inst_base
from simbricks.orchestration.system import host as sys_host
from simbricks.orchestration.system import pcie as sys_pcie
from simbricks.orchestration.system import mem as sys_mem
from simbricks.utils import base as utils_base, file as util_file
from simbricks.orchestration.instantiation import socket as inst_socket


class HostSim(sim_base.Simulator):

    def __init__(self, simulation: sim_base.Simulation, executable: str, name=""):
        super().__init__(simulation=simulation, executable=executable, name=name)

    def toJSON(self) -> dict:
        return super().toJSON()

    @classmethod
    def fromJSON(cls, simulation: sim_base.Simulation, json_obj: dict) -> Gem5Sim:
        return super().fromJSON(simulation, json_obj)

    def full_name(self) -> str:
        return "host." + self.name

    def add(self, host: sys_host.Host):
        super().add(host)

    def config_str(self) -> str:
        return []

    def supported_image_formats(self) -> list[str]:
        raise Exception("implement me")

    def supported_socket_types(
        self, interface: system.Interface
    ) -> set[inst_socket.SockType]:
        return {inst_socket.SockType.CONNECT}


class Gem5Sim(HostSim):

    def __init__(self, simulation: sim_base.Simulation):
        super().__init__(
            simulation=simulation, executable="sims/external/gem5/build/X86/gem5"
        )
        self.name = f"Gem5Sim-{self._id}"
        self.cpu_type_cp = "X86KvmCPU"
        self.cpu_type = "TimingSimpleCPU"
        self.extra_main_args: list[str] = []
        self.extra_config_args: list[str] = []
        self._variant: str = "fast"
        self._sys_clock: str = "1GHz"  # TODO: move to system module

    def supports_checkpointing(self) -> bool:
        return True

    def resreq_cores(self) -> int:
        return 1

    def resreq_mem(self) -> int:
        return 1024

    def supported_image_formats(self) -> list[str]:
        return ["raw"]

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        json_obj["cpu_type_cp"] = self.cpu_type_cp
        json_obj["cpu_type"] = self.cpu_type
        json_obj["extra_main_args"] = self.extra_main_args
        json_obj["extra_config_args"] = self.extra_config_args
        json_obj["_variant"] = self._variant
        json_obj["_sys_clock"] = self._sys_clock
        return json_obj

    @classmethod
    def fromJSON(cls, simulation: sim_base.Simulation, json_obj: dict) -> Gem5Sim:
        instance = super().fromJSON(simulation, json_obj)
        instance.cpu_type_cp = utils_base.get_json_attr_top(json_obj, "cpu_type_cp")
        instance.cpu_type = utils_base.get_json_attr_top(json_obj, "cpu_type")
        instance.extra_main_args = utils_base.get_json_attr_top(
            json_obj, "extra_main_args"
        )
        instance.extra_config_args = utils_base.get_json_attr_top(
            json_obj, "extra_config_args"
        )
        instance._variant = utils_base.get_json_attr_top(json_obj, "_variant")
        instance._sys_clock = utils_base.get_json_attr_top(json_obj, "_sys_clock")
        return instance

    async def prepare(self, inst: inst_base.Instantiation) -> None:
        await super().prepare(inst=inst)
        util_file.mkdir(inst.cpdir_subdir(sim=self))

    def checkpoint_commands(self) -> list[str]:
        return ["m5 checkpoint"]

    def cleanup_commands(self) -> list[str]:
        return ["m5 exit"]

    def run_cmd(self, inst: inst_base.Instantiation) -> str:
        cpu_type = self.cpu_type
        if inst.create_checkpoint:
            cpu_type = self.cpu_type_cp

        full_sys_hosts = self.filter_components_by_type(ty=sys_host.BaseLinuxHost)
        if len(full_sys_hosts) != 1:
            raise Exception("Gem5Sim only supports simulating 1 FullSystemHost")
        host_spec = full_sys_hosts[0]

        cmd = f"{inst.join_repo_base(f'{self._executable}.{self._variant}')} --outdir={inst.get_simmulator_output_dir(sim=self)} "
        cmd += " ".join(self.extra_main_args)
        cmd += (
            f" {inst.join_repo_base('sims/external/gem5/configs/simbricks/simbricks.py')} --caches --l2cache "
            "--l1d_size=32kB --l1i_size=32kB --l2_size=32MB "
            "--l1d_assoc=8 --l1i_assoc=8 --l2_assoc=16 "
            f"--cacheline_size=64 --cpu-clock={host_spec.cpu_freq}"
            f" --sys-clock={self._sys_clock} "
            f"--checkpoint-dir={inst.cpdir_subdir(sim=self)} "
            f"--kernel={inst.join_repo_base('images/vmlinux')} "
        )
        for disk in host_spec.disks:
            cmd += f"--disk-image={disk.path(inst=inst, format='raw')} "
        cmd += (
            f"--cpu-type={cpu_type} --mem-size={host_spec.memory}MB "
            f"--num-cpus={host_spec.cores} "
            "--mem-type=DDR4_2400_16x4 "
        )

        if host_spec.kcmd_append is not None:
            cmd += f'--command-line-append="{host_spec.kcmd_append}" '

        if inst.create_checkpoint:
            cmd += "--max-checkpoints=1 "

        if inst.restore_checkpoint:
            cmd += "-r 1 "

        latency, sync_period, run_sync = (
            sim_base.Simulator.get_unique_latency_period_sync(
                channels=self.get_channels()
            )
        )

        fsh_interfaces = host_spec.interfaces()

        pci_interfaces = system.Interface.filter_by_type(
            interfaces=fsh_interfaces, ty=sys_pcie.PCIeHostInterface
        )
        for inf in pci_interfaces:
            socket = inst.update_get_socket(interface=inf)
            if socket is None:
                continue
            assert socket._type == inst_socket.SockType.CONNECT
            cmd += (
                f"--simbricks-pci=connect:{socket._path}"
                f":latency={latency}ns"
                f":sync_interval={sync_period}ns"
            )
            if run_sync and not inst.create_checkpoint:
                cmd += ":sync"
            cmd += " "

        mem_interfaces = system.Interface.filter_by_type(
            interfaces=fsh_interfaces, ty=sys_mem.MemHostInterface
        )
        for inf in mem_interfaces:
            socket = inst.update_get_socket(interface=inf)
            if socket is None:
                continue
            assert socket._type == inst_socket.SockType.CONNECT
            utils_base.has_expected_type(inf.component, sys_mem.MemSimpleDevice)
            dev: sys_mem.MemSimpleDevice = inf.component
            cmd += (
                f"--simbricks-mem={dev._size}@{dev._addr}@{dev._as_id}@"
                f"connect:{socket._path}"
                f":latency={latency}ns"
                f":sync_interval={sync_period}ns"
            )
            if run_sync and not inst.create_checkpoint:
                cmd += ":sync"
            cmd += " "

        # TODO: FIXME
        # for net in self.net_directs:
        #     cmd += (
        #         '--simbricks-eth-e1000=listen'
        #         f':{env.net2host_eth_path(net, self)}'
        #         f':{env.net2host_shm_path(net, self)}'
        #         f':latency={net.eth_latency}ns'
        #         f':sync_interval={net.sync_period}ns'
        #     )
        #     if cpu_type == 'TimingSimpleCPU':
        #         cmd += ':sync'
        #     cmd += ' '

        cmd += " ".join(self.extra_config_args)

        return cmd


class QemuSim(HostSim):

    def __init__(self, simulation: sim_base.Simulation) -> None:
        super().__init__(
            simulation=simulation,
            executable="sims/external/qemu/build/x86_64-softmmu/qemu-system-x86_64",
        )
        self.name = f"QemuSim-{self._id}"
        self._disks: list[tuple[str, str]] = []  # [(path, format)]

    def resreq_cores(self) -> int:
        return 1

    def resreq_mem(self) -> int:
        return 1024

    def supported_image_formats(self) -> list[str]:
        return ["raw", "qcow"]

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        # disks is created upon invocation of "prepare", hence we do not need to serialize it
        return json_obj

    @classmethod
    def fromJSON(cls, simulation: sim_base.Simulation, json_obj: dict) -> QemuSim:
        return super().fromJSON(simulation, json_obj)

    async def prepare(self, inst: inst_base.Instantiation) -> None:
        await super().prepare(inst=inst)

        full_sys_hosts = self.filter_components_by_type(ty=sys_host.FullSystemHost)
        if len(full_sys_hosts) != 1:
            raise Exception("QEMU only supports simulating 1 FullSystemHost")

        d = []
        for disk in full_sys_hosts[0].disks:
            format = (
                "qcow2"
                if not isinstance(disk, sys_host.LinuxConfigDiskImage)
                else "raw"
            )
            copy_path = await disk.make_qcow_copy(
                inst=inst,
                format=format,
                sim=self
            )
            assert copy_path is not None
            d.append((copy_path, format))
        self._disks = d

    def checkpoint_commands(self) -> list[str]:
        return []

    def cleanup_commands(self) -> list[str]:
        return ["poweroff -f"]

    def run_cmd(self, inst: inst_base.Instantiation) -> str:

        latency, period, sync = sim_base.Simulator.get_unique_latency_period_sync(
            channels=self.get_channels()
        )
        accel = ",accel=kvm:tcg" if not sync else ""

        cmd = (
            f"{inst.join_repo_base(relative_path=self._executable)} -machine q35{accel} -serial mon:stdio "
            "-cpu Skylake-Server -display none -nic none "
            f"-kernel {inst.join_repo_base('images/bzImage')} "
        )

        full_sys_hosts = self.filter_components_by_type(ty=sys_host.BaseLinuxHost)
        if len(full_sys_hosts) != 1:
            raise Exception("QEMU only supports simulating 1 FullSystemHost")
        host_spec = full_sys_hosts[0]

        kcmd_append = ""
        if host_spec.kcmd_append is not None:
            kcmd_append = " " + host_spec.kcmd_append

        for index, disk in enumerate(self._disks):
            cmd += f"-drive file={disk[0]},if=ide,index={index},media=disk,driver={disk[1]} "
        cmd += (
            '-append "earlyprintk=ttyS0 console=ttyS0 root=/dev/sda1 '
            f'init=/home/ubuntu/guestinit.sh rw{kcmd_append}" '
            f"-m {host_spec.memory} -smp {host_spec.cores} "
        )

        if sync:
            unit = host_spec.cpu_freq[-3:]
            if unit.lower() == "ghz":
                base = 0
            elif unit.lower() == "mhz":
                base = 3
            else:
                raise ValueError("cpu frequency specified in unsupported unit")
            num = float(host_spec.cpu_freq[:-3])
            shift = base - int(math.ceil(math.log(num, 2)))

            cmd += f" -icount shift={shift},sleep=off "

        fsh_interfaces = host_spec.interfaces()
        pci_interfaces = system.Interface.filter_by_type(
            interfaces=fsh_interfaces, ty=sys_pcie.PCIeHostInterface
        )
        for inf in pci_interfaces:
            socket = inst.update_get_socket(interface=inf)
            if socket is None:
                continue
            assert socket._type is inst_socket.SockType.CONNECT
            cmd += f"-device simbricks-pci,socket={socket._path}"
            if sync:
                cmd += ",sync=on"
                cmd += f",pci-latency={latency}"
                cmd += f",sync-period={period}"
            else:
                cmd += ",sync=off"
            cmd += " "

        return cmd



# Copyright 2024 Max Planck Institute for Software Systems, and
# National University of Singapore
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from simbricks.orchestration import system
from simbricks.orchestration.simulation import base as sim_base
from simbricks.orchestration.simulation import channel as sim_chan
from simbricks.utils import base as utils_base


def add_specs(simulator: sim_base.Simulator, *specifications) -> None:
    utils_base.has_expected_type(obj=simulator, expected_type=sim_base.Simulator)
    for spec in specifications:
        utils_base.has_expected_type(obj=spec, expected_type=system.Component)
        simulator.add(comp=spec)


def enable_sync_simulation(
    simulation: sim_base.Simulation, amount: int | None = None, ratio: sim_chan.Time = None
) -> None:
    utils_base.has_expected_type(obj=simulation, expected_type=sim_base.Simulation)
    set_period: bool = amount is not None and ratio is not None
    if set_period:
        utils_base.has_expected_type(obj=amount, expected_type=int)
        utils_base.has_expected_type(obj=ratio, expected_type=sim_chan.Time)

    for chan in simulation.get_all_channels():
        chan._synchronized = True
        if set_period:
            chan.set_sync_period(amount=amount, ratio=ratio)


def disalbe_sync_simulation(simulation: sim_base.Simulation) -> None:
    utils_base.has_expected_type(obj=simulation, expected_type=sim_base.Simulation)

    for chan in simulation.get_all_channels(lazy=False):
        chan._synchronized = False


def simple_simulation(
    system: system.System,
    sync=False,
    compmap=dict[type[system.Component], type[sim_base.Simulator]]
):
  """Create simple simulation from system. Uses a map from component type to
  simulator type and then creates one simulator per component."""
  # FIXME: name from system, but system has no name
  simulation = sim_base.Simulation(
      name="netperf_sysconf", system=system
  )

  for comp in system._all_components.values():
    if comp in simulation._sys_sim_map:
      continue

    for (ct,st) in compmap.items():
      if isinstance(comp, ct):
        simulator = st(simulation)
        simulator.add(comp)
        if comp.name:
          simulator.name = comp.name

    if not sync:
      disalbe_sync_simulation(simulation=simulation)

  return simulation


# Copyright 2024 Max Planck Institute for Software Systems, and
# National University of Singapore
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import annotations

from simbricks.orchestration.system import base
from simbricks.utils import base as utils_base


class PCIeHostInterface(base.Interface):
    # Note AK: Component here is on purpose. Other simulators than host
    # simulators can also have PCIeHost interfaces (e.g. a PCIe switch)
    def __init__(self, c: base.Component) -> None:
        super().__init__(c)


class PCIeDeviceInterface(base.Interface):
    def __init__(self, c: base.Component) -> None:
        super().__init__(c)

    def connect(self, c: base.Channel) -> None:
        # Note AK: a bit ugly, but I think we can't get around a rt check here
        utils_base.has_expected_type(c, PCIeChannel)
        super().connect(c)


class PCIeChannel(base.Channel):
    def __init__(self, host: PCIeHostInterface, dev: PCIeDeviceInterface) -> None:
        super().__init__(host, dev)

    def host_if(self) -> PCIeHostInterface:
        return self.a

    def dev_if(self) -> PCIeDeviceInterface:
        return self.b


# Note AK: Can we make this abstract?
class PCIeSimpleDevice(base.Component):
    def __init__(self, s: base.System):
        super().__init__(s)
        self._pci_if: PCIeDeviceInterface = PCIeDeviceInterface(self)
        super().add_if(self._pci_if)

    def add_if(self, interface: PCIeDeviceInterface) -> None:
        utils_base.has_expected_type(interface, PCIeDeviceInterface)
        if self._pci_if:
            raise Exception(
                f"you overwrite PCIeSimpleDevice._pci_if. ({self._pci_if.id()} -> {interface.id()})"
            )
        self._pci_if = interface
        super().add_if(interface)

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        json_obj["pci_if"] = self._pci_if.id()
        return json_obj

    @classmethod
    def fromJSON(cls, system: base.System, json_obj: dict) -> PCIeSimpleDevice:
        instance = super().fromJSON(system=system, json_obj=json_obj)
        inf_id = int(utils_base.get_json_attr_top(json_obj, "pci_if"))
        instance._pci_if = system.get_inf(inf_id)
        return instance



# Copyright 2024 Max Planck Institute for Software Systems, and
# National University of Singapore
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import annotations

import pathlib
from simbricks.orchestration import system
from simbricks.orchestration.system import base as sys_base
from simbricks.orchestration.system import eth as sys_eth
from simbricks.orchestration.simulation import base as sim_base
from simbricks.orchestration.simulation.net import ns3_components as ns3_comps
from simbricks.orchestration.instantiation import base as inst_base
from simbricks.orchestration.instantiation import socket as inst_socket
from simbricks.utils import base as base_utils


class NetSim(sim_base.Simulator):
    """Base class for network simulators."""

    def __init__(
        self,
        simulation: sim_base.Simulation,
        executable: str,
        name: str = "",
    ) -> None:
        super().__init__(simulation=simulation, executable=executable, name=name)

    def full_name(self) -> str:
        return "net." + self.name

    def init_network(self) -> None:
        pass

    def supported_socket_types(
        self, interface: sys_base.Interface
    ) -> set[inst_socket.SockType]:
        return {inst_socket.SockType.CONNECT}

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        return json_obj

    @classmethod
    def fromJSON(cls, simulation: sim_base.Simulation, json_obj: dict) -> NetSim:
        return super().fromJSON(simulation, json_obj)


class WireNet(NetSim):

    def __init__(
        self, simulation: sim_base.Simulation, relative_pcap_filepath: str | None = None
    ) -> None:
        super().__init__(
            simulation=simulation,
            executable="sims/net/wire/net_wire",
        )
        self.name = f"WireNet-{self._id}"
        self._relative_pcap_file_path: str | None = relative_pcap_filepath

    def add(self, wire: sys_eth.EthWire):
        base_utils.has_expected_type(wire, sys_eth.EthWire)
        if len(self._components) > 1:
            raise Exception(
                "can only add a single wire component to the WireNet simulator"
            )
        super().add(wire)

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        json_obj["relative_pcap_file_path"] = self._relative_pcap_file_path
        return json_obj

    @classmethod
    def fromJSON(cls, simulation: sim_base.Simulation, json_obj: dict) -> WireNet:
        instance = super().fromJSON(simulation, json_obj)
        instance._relative_pcap_file_path = base_utils.get_json_attr_top(json_obj, "relative_pcap_file_path") 
        return instance

    def run_cmd(self, inst: inst_base.Instantiation) -> str:
        channels = self.get_channels()
        eth_latency, sync_period, run_sync = (
            sim_base.Simulator.get_unique_latency_period_sync(channels=channels)
        )

        sockets = self._get_socks_by_all_comp(inst=inst)
        assert len(sockets) == 2

        cmd = inst.join_repo_base(self._executable)
        cmd += f"{sockets[0]._path} {sockets[1]._path} {run_sync} {sync_period} {eth_latency}"

        if self._relative_pcap_file_path is not None:
            pcap_file = inst.join_output_base(
                relative_path=self._relative_pcap_file_path
            )
            cmd += " " + pcap_file
        return cmd


class SwitchNet(NetSim):

    def __init__(
        self,
        simulation: sim_base.Simulation,
        executable: str = "sims/net/switch/net_switch",
        relative_pcap_filepath: str | None = None,
    ) -> None:
        super().__init__(
            simulation=simulation,
            executable=executable,
        )
        self.name = f"SwitchNet-{self._id}"
        self._relative_pcap_file_path: str | None = relative_pcap_filepath

    def add(self, switch_spec: sys_eth.EthSwitch):
        base_utils.has_expected_type(switch_spec, sys_eth.EthSwitch)
        if len(self._components) > 1:
            raise Exception("can only add a single switch component to the SwitchNet")
        super().add(switch_spec)

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        json_obj["relative_pcap_file_path"] = self._relative_pcap_file_path
        return json_obj

    @classmethod
    def fromJSON(cls, simulation: sim_base.Simulation, json_obj: dict) -> SwitchNet:
        instance = super().fromJSON(simulation, json_obj)
        instance._relative_pcap_file_path = base_utils.get_json_attr_top(json_obj, "relative_pcap_file_path") 
        return instance

    def run_cmd(self, inst: inst_base.Instantiation) -> str:
        channels = self.get_channels()
        eth_latency, sync_period, run_sync = (
            sim_base.Simulator.get_unique_latency_period_sync(channels=channels)
        )

        cmd = inst.join_repo_base(self._executable)
        cmd += f" -S {sync_period} -E {eth_latency}"

        if not run_sync:
            cmd += " -u"

        if self._relative_pcap_file_path is not None:
            pcap_file = inst.join_output_base(
                relative_path=self._relative_pcap_file_path
            )
            cmd += " " + pcap_file

        sockets = self._get_socks_by_all_comp(inst=inst)
        listen, connect = sim_base.Simulator.split_sockets_by_type(sockets)

        for sock in connect:
            cmd += " -s " + sock._path

        for sock in listen:
            cmd += " -h " + sock._path

        return cmd


class MemSwitchNet(SwitchNet):

    def __init__(
        self, simulation: sim_base.Simulation, relative_pcap_file_path=None
    ) -> None:
        super().__init__(
            simulation=simulation,
            executable="sims/mem/memswitch/memswitch",
            relative_pcap_file_path=relative_pcap_file_path,
        )
        self.name = f"MemSwitchNet-{self._id}"
        """AS_ID,VADDR_START,VADDR_END,MEMNODE_MAC,PHYS_START."""
        self.mem_map = []

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        # TODO: FIXME
        return json_obj

    @classmethod
    def fromJSON(cls, simulation: sim_base.Simulation, json_obj: dict) -> MemSwitchNet:
        instance = super().fromJSON(simulation, json_obj)
        # TODO: FIXME
        return instance

    def run_cmd(self, inst: inst_base.Instantiation) -> str:
        cmd = super().run_cmd(inst)

        for m in self.mem_map:
            cmd += " -m " + f" {m[0]},{m[1]},{m[2]},"
            cmd += "".join(reversed(m[3].split(":")))
            cmd += f",{m[4]}"

        return cmd


class SimpleNS3Sim(NetSim):

    def __init__(
        self,
        simulation: sim_base.Simulation,
        name: str = "SimpleNS3Sim",
        ns3_run_script: str = "",
    ) -> None:
        super().__init__(
            simulation=simulation,
            executable="sims/external/ns-3/simbricks-run.sh",
            name=name,
        )
        self._ns3_run_script: str = ns3_run_script
        self.opt: str | None = None

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        json_obj["ns3_run_script"] = self._ns3_run_script
        json_obj["opt"] = self.opt
        return json_obj

    @classmethod
    def fromJSON(cls, simulation: sim_base.Simulation, json_obj: dict) -> SimpleNS3Sim:
        instance = super().fromJSON(simulation, json_obj)
        instance._ns3_run_script = base_utils.get_json_attr_top(
            json_obj, "ns3_run_script"
        )
        instance.opt = base_utils.get_json_attr_top_or_none(json_obj, "opt")
        return instance

    def run_cmd(self, inst: inst_base.Instantiation) -> str:
        return f"{inst.join_repo_base(self._executable)} {self._ns3_run_script} "


class NS3DumbbellNet(SimpleNS3Sim):

    def __init__(self, simulation: sim_base.Simulation) -> None:
        super().__init__(
            simulation=simulation,
            ns3_run_script="simbricks-dumbbell-example",
        )
        self.name = f"NS3DumbbellNet-{self._id}"
        self._left: sys_eth.EthSwitch | None = None
        self._right: sys_eth.EthSwitch | None = None

    def add(self, left: sys_eth.EthSwitch, right: sys_eth.EthSwitch):
        base_utils.has_expected_type(left, sys_eth.EthSwitch)
        base_utils.has_expected_type(right, sys_eth.EthSwitch)

        if (
            len(self._components) > 2
            or self._left is not None
            or self._right is not None
        ):
            raise Exception("NS3DumbbellNet can only simulate two switches")

        super().add(comp=left)
        super().add(comp=right)
        self._left = left
        self._right = right

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        json_obj["left"] = self._left.id()
        json_obj["right"] = self._right.id()
        return json_obj

    @classmethod
    def fromJSON(
        cls, simulation: sim_base.Simulation, json_obj: dict
    ) -> NS3DumbbellNet:
        instance = super().fromJSON(simulation, json_obj)
        left_id = int(base_utils.get_json_attr_top(json_obj, "left"))
        instance._left = json_obj["left"] = simulation.system.get_comp(left_id)
        right_id = int(base_utils.get_json_attr_top(json_obj, "right"))
        instance._right = json_obj["right"] = simulation.system.get_comp(right_id)
        return instance

    def run_cmd(self, inst: inst_base.Instantiation) -> str:
        cmd = super().run_cmd(inst=inst)

        left_socks = self._get_socks_by_comp(inst=inst, comp=self._left)
        for sock in left_socks:
            assert sock._type == inst_socket.SockType.CONNECT
            cmd += f"--SimbricksPortLeft={sock._path} "

        right_sockets = self._get_socks_by_comp(inst=inst, comp=self._right)
        for sock in right_sockets:
            assert sock._type == inst_socket.SockType.CONNECT
            cmd += f"--SimbricksPortRight={sock._path} "

        if self.opt is not None:
            cmd += f"{self.opt}"

        print(cmd)
        return cmd


class NS3BridgeNet(SimpleNS3Sim):

    def __init__(self, simulation: sim_base.Simulation) -> None:
        super().__init__(
            simulation=simulation,
            ns3_run_script="simbricks-bridge-example",
        )
        self.name = f"NS3BridgeNet-{self._id}"

    def add(self, switch_comp: sys_eth.EthSwitch):
        base_utils.has_expected_type(switch_comp, sys_eth.EthSwitch)
        if len(self._components) > 1:
            raise Exception("NS3BridgeNet can only simulate one switch/bridge")
        super().add(comp=switch_comp)

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        return json_obj

    @classmethod
    def fromJSON(cls, simulation: sim_base.Simulation, json_obj: dict) -> NS3BridgeNet:
        instance = super().fromJSON(simulation, json_obj)
        # TODO: FIXME
        return instance

    def run_cmd(self, inst: inst_base.Instantiation) -> str:
        cmd = super().run_cmd(inst=inst)

        sockets = self._get_socks_by_all_comp(inst=inst)
        for sock in sockets:
            cmd += f"--SimbricksPort={sock._path} "

        if self.opt is not None:
            cmd += f"{self.opt}"

        return cmd


class NS3Net(SimpleNS3Sim):

    def __init__(self, simulation: sim_base.Simulation):
        super().__init__(simulation, ns3_run_script="e2e-cc-example")
        self.name = f"NS3Net-{self._id}"
        self.use_file = True
        self.global_conf = ns3_comps.NS3GlobalConfig()
        self.logging = ns3_comps.NS3Logging()

    def add(self, comp: sys_base.Component):
        super().add(comp)

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        json_obj["use_file"] = self.use_file
        json_obj["global_conf"] = self.global_conf.toJSON()
        json_obj["logging"] = self.logging.toJSON()
        return json_obj

    @classmethod
    def fromJSON(cls,
                 simulation: sim_base.Simulation,
                 json_obj: dict) -> NS3Net:
        instance = super().fromJSON(simulation, json_obj)
        instance.use_file = base_utils.get_json_attr_top(json_obj, "use_file")
        instance.global_conf = ns3_comps.NS3GlobalConfig.fromJSON(
            base_utils.get_json_attr_top(json_obj, "global_conf")
        )
        instance.logging = ns3_comps.NS3Logging.fromJSON(
            base_utils.get_json_attr_top(json_obj, "logging")
        )
        return instance

    def supported_socket_types(
        self, interface: sys_base.Interface
    ) -> set[inst_socket.SockType]:
        return [inst_socket.SockType.CONNECT, inst_socket.SockType.LISTEN]


    def run_cmd(self, inst: inst_base.Instantiation) -> str:
        cmd = super().run_cmd(inst=inst)

        ns3_components: map[sys_base.Component, ns3_comps.NS3Component] = {}
        ns3c: set[ns3_comps.NS3Component] = set()

        # TODO: with the current abstraction we connect hosts directly to
        # switches in ns-3 without specifying any NICs explicitly. We should
        # change this in the future.

        # create the ns3 components
        for comp in self.components():
            if isinstance(comp, sys_eth.EthSwitch):
                ns3_switch = ns3_comps.NS3SwitchNode(comp)
                ns3_components[comp] = ns3_switch
                ns3c.add(ns3_switch)
            elif isinstance(comp, system.Host):
                ns3_host = ns3_comps.NS3SimpleHost(comp)
                ns3_components[comp] = ns3_host
                for app in comp.applications:
                    ns3_app = ns3_comps.NS3GenericApplication(app)
                    ns3_components[app] = ns3_app
                    ns3_host.add_component(ns3_app)

        def get_opposing_component(chan: sys_base.Channel,
                                   comp: sys_base.Component
                                   ) -> sys_base.Component:
            if chan.a.component == comp:
                return chan.b.component
            assert(chan.b.component == comp)
            return chan.a.component

        def get_component_interface(chan: sys_base.Channel,
                                    comp: sys_base.Component
                                    ) -> sys_base.Interface:
            if chan.a.component == comp:
                return chan.a
            else:
                return chan.b

        # create the correct channels, i.e. iterate over all channels and figure
        # out for each of them how to realize them in ns-3
        channels: set[sys_base.Channel] = set()

        for comp in self.components():
            for chan in comp.channels():
                if chan in channels:
                    continue

                assert(isinstance(chan, sys_eth.EthChannel))
                other = get_opposing_component(chan, comp)

                if isinstance(comp, system.Host):
                    # connect host to switch
                    assert(isinstance(other, sys_eth.EthSwitch))
                    ns3_components[other].add_component(ns3_components[comp])
                elif isinstance(comp, sys_eth.EthSwitch):
                    if other not in self.components():
                        # the component is in a different simulator instance
                        sim_chan = self._simulation.retrieve_or_create_channel(chan)
                        socket = inst.update_get_socket(get_component_interface(chan, comp))
                        assert(socket)
                        ns3_sb_chan = ns3_comps.NS3NetworkSimbricks(other, sim_chan, socket)
                        ns3_components[comp].add_component(ns3_sb_chan)
                    else:
                        if isinstance(other, system.Host):
                            # we handle this case when we see the host
                            continue
                        # two switches in this ns-3 instance
                        assert(isinstance(other, sys_eth.EthSwitch))
                        ns3_chan = ns3_comps.NS3SimpleChannel(chan)
                        ns3_chan.left_node = ns3_components[comp]
                        ns3_chan.right_node = ns3_components[other]
                        ns3c.add(ns3_chan)
                else:
                    raise ValueError("Cannot add component to ns-3")
                
                channels.add(chan)

        for component in ns3c:
            component.resolve_paths()

        params: list[str] = []
        params.append(self.global_conf.ns3_config())
        params.append(self.logging.ns3_config())
        for component in ns3c:
            params.append(component.ns3_config())

        #params.append(" ".join([f"--{k}={v}" for k,v in self.opts.items()]))
        if self.opt:
            params.append(self.opt)

        params_str = "\n".join(params)

        if self.use_file:
            # TODO: change this to a more sensible file path?
            sim_out = inst.get_simmulator_output_dir(self)
            pathlib.Path(sim_out).mkdir(parents=True, exist_ok=True)
            file_path = inst._join_paths(sim_out, f"{self.name}_params")
            with open(file_path, 'w', encoding="utf-8") as f:
                f.write(params_str)
            cmd += f"--ConfigFile={file_path}"
        else:
            cmd += params_str

        return cmd



# Copyright 2024 Max Planck Institute for Software Systems, and
# National University of Singapore
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import annotations

import abc
import io
import pathlib
import tarfile
import typing as tp
from simbricks.utils import base as utils_base
from simbricks.orchestration.instantiation import base as inst_base
from simbricks.orchestration.system import base as sys_base

if tp.TYPE_CHECKING:
    from simbricks.orchestration.system import host as sys_host
    from simbricks.orchestration.simulation import base as sim_base


class DiskImage(utils_base.IdObj):
    def __init__(self, h: sys_host.Host) -> None:
        super().__init__()
        self.host: sys_host.Host = h
        self._qemu_img_exec: str = "sims/external/qemu/build/qemu-img"

    @abc.abstractmethod
    def available_formats(self) -> list[str]:
        return []

    @abc.abstractmethod
    def path(self, inst: inst_base.Instantiation, format: str) -> str:
        raise Exception("must be overwritten")

    async def make_qcow_copy(
        self,
        inst: inst_base.Instantiation,
        format: str,
        sim: sim_base.Simulator
    ) -> str:
        disk_path = pathlib.Path(self.path(inst=inst, format=format))
        copy_path = inst.join_imgs_path(relative_path=f"hdcopy.{self._id}")
        prep_cmds = [
            (
                f"{inst.join_repo_base(relative_path=self._qemu_img_exec)} create -f qcow2 -F qcow2 -o "
                f'backing_file="{disk_path}" '
                f"{copy_path}"
            )
        ]
        await inst._cmd_executor.exec_simulator_prepare_cmds(sim, prep_cmds)
        return copy_path

    @staticmethod
    def assert_is_file(path: str) -> str:
        if not pathlib.Path(path).is_file():
            raise Exception(f"path={path} must be a file")

    async def _prepare_format(self, inst: inst_base.Instantiation, format: str) -> None:
        pass

    async def prepare(self, inst: inst_base.Instantiation) -> None:
        # Find first supported disk image format in order of simulator pref.
        sim = inst.find_sim_by_spec(self.host)
        format = None
        av_fmt = self.available_formats()
        for f in sim.supported_image_formats():
            if f in av_fmt:
                format = f
                break

        if format is None:
            raise Exception("No supported image format found")

        await self._prepare_format(inst, format)

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        json_obj["host"] = self.host.id()
        json_obj["qemu_img_exec"] = self._qemu_img_exec
        return json_obj

    @classmethod
    def fromJSON(cls, system: sys_base.System, json_obj: dict) -> DiskImage:
        instance = super().fromJSON(json_obj)
        host_id = int(utils_base.get_json_attr_top(json_obj, "host"))
        instance.host = system.get_comp(host_id)
        instance._qemu_img_exec = utils_base.get_json_attr_top(
            json_obj, "qemu_img_exec"
        )
        return instance


# Disk image where user just provides a path
class ExternalDiskImage(DiskImage):
    def __init__(self, h: sys_host.FullSystemHost, path: str) -> None:
        super().__init__(h)
        self._path = path
        self.formats = ["raw", "qcow2"]

    def available_formats(self) -> list[str]:
        return self.formats

    def path(self, inst: inst_base.Instantiation, format: str) -> str:
        DiskImage.assert_is_file(self._path)
        return self._path

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        json_obj["path"] = self._path
        json_obj["formats"] = self.formats
        return json_obj

    @classmethod
    def fromJSON(cls, system: sys_base.System, json_obj: dict) -> DiskImage:
        instance = super().fromJSON(system, json_obj)
        instance._path = utils_base.get_json_attr_top(json_obj, "path")
        instance.formats = utils_base.get_json_attr_top(json_obj, "formats")
        return instance


# Disk images shipped with simbricks
class DistroDiskImage(DiskImage):
    def __init__(self, h: sys_host.FullSystemHost, name: str) -> None:
        super().__init__(h)
        self.name = name
        self.formats = ["raw", "qcow2"]

    def available_formats(self) -> list[str]:
        return self.formats

    def path(self, inst: inst_base.Instantiation, format: str) -> str:
        path = inst.hd_path(self.name)
        if format == "raw":
            path += ".raw"
        elif format == "qcow2":
            pass
        else:
            raise RuntimeError("Unsupported disk format")
        DiskImage.assert_is_file(path)
        return path

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        json_obj["name"] = self.name
        json_obj["formats"] = self.formats
        return json_obj

    @classmethod
    def fromJSON(cls, system: sys_base.System, json_obj: dict) -> DiskImage:
        instance = super().fromJSON(system, json_obj)
        instance.name = utils_base.get_json_attr_top(json_obj, "name")
        instance.formats = utils_base.get_json_attr_top(json_obj, "formats")
        return instance


# Abstract base class for dynamically generated images
class DynamicDiskImage(DiskImage):
    def __init__(self, h: sys_host.FullSystemHost) -> None:
        super().__init__(h)

    def path(self, inst: inst_base.Instantiation, format: str) -> str:
        return inst.dynamic_img_path(self, format)

    @abc.abstractmethod
    async def _prepare_format(self, inst: inst_base.Instantiation, format: str) -> None:
        pass

    @classmethod
    def fromJSON(cls, system: sys_base.System, json_obj: dict) -> DynamicDiskImage:
        return super().fromJSON(system, json_obj)


# Builds the Tar with the commands to run etc.
class LinuxConfigDiskImage(DynamicDiskImage):
    def __init__(self, h: sys_host.LinuxHost) -> None:
        super().__init__(h)
        self.host: sys_host.LinuxHost

    def available_formats(self) -> list[str]:
        return ["raw"]

    async def make_qcow_copy(
        self, inst: inst_base.Instantiation, format: str, sim: sim_base.Simulator
    ) -> str:
        return self.path(inst=inst, format=format)

    async def _prepare_format(self, inst: inst_base.Instantiation, format: str) -> None:
        path = self.path(inst, format)
        with tarfile.open(path, "w:") as tar:
            # add main run script
            cfg_i = tarfile.TarInfo("guest/run.sh")
            cfg_i.mode = 0o777
            cfg_f = self.host.strfile(self.host.config_str(inst))
            cfg_f.seek(0, io.SEEK_END)
            cfg_i.size = cfg_f.tell()
            cfg_f.seek(0, io.SEEK_SET)
            tar.addfile(tarinfo=cfg_i, fileobj=cfg_f)
            cfg_f.close()

            # add additional config files
            for n, f in self.host.config_files(inst).items():
                f_i = tarfile.TarInfo("guest/" + n)
                f_i.mode = 0o777
                f.seek(0, io.SEEK_END)
                f_i.size = f.tell()
                f.seek(0, io.SEEK_SET)
                tar.addfile(tarinfo=f_i, fileobj=f)
                f.close()

    @classmethod
    def fromJSON(cls, system: sys_base.System, json_obj: dict) -> LinuxConfigDiskImage:
        return super().fromJSON(system, json_obj)


# This is an additional example: building disk images directly from python
# Could of course also have a version that generates the packer config from
# python
class PackerDiskImage(DynamicDiskImage):
    def __init__(self, h: sys_host.FullSystemHost, packer_config_path: str) -> None:
        super().__init__(h)
        self.config_path = packer_config_path

    def available_formats(self) -> list[str]:
        return ["raw", "qcow"]

    async def _prepare_format(self, inst: inst_base.Instantiation, format: str) -> None:
        # TODO: invoke packer to build the image if necessary
        pass

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        json_obj["config_path"] = self.config_path
        return json_obj

    @classmethod
    def fromJSON(cls, system: sys_base.System, json_obj: dict) -> PackerDiskImage:
        instance = super().fromJSON(system, json_obj)
        instance.config_path = utils_base.get_json_attr_top(json_obj, "config_path")
        return instance



# Copyright 2024 Max Planck Institute for Software Systems, and
# National University of Singapore
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import annotations

from simbricks.orchestration.system import base as sys_base
from simbricks.orchestration.system import pcie as sys_pcie
from simbricks.orchestration.system import eth as sys_eth
from simbricks.orchestration.system import nic as sys_nic
from simbricks.orchestration.instantiation import base as inst_base
from simbricks.orchestration.instantiation import socket as inst_socket
from simbricks.orchestration.simulation import base as sim_base


class PCIDevSim(sim_base.Simulator):
    """Base class for PCIe device simulators."""

    def __init__(self, simulation: sim_base.Simulation, executable: str, name: str) -> None:
        super().__init__(simulation=simulation, executable=executable, name=name)

    def full_name(self) -> str:
        return "dev." + self.name

    def supported_socket_types(self, interface: sys_base.Interface) -> set[inst_socket.SockType]:
        return {inst_socket.SockType.LISTEN}


class NICSim(PCIDevSim):
    """Base class for NIC simulators."""

    def __init__(self, simulation: sim_base.Simulation, executable: str, name: str = "") -> None:
        super().__init__(simulation=simulation, executable=executable, name=name)

    def full_name(self) -> str:
        return "nic." + self.name

    def add(self, nic: sys_nic.SimplePCIeNIC):
        assert len(self._components) < 1
        super().add(nic)

    def run_cmd(self, inst: inst_base.Instantiation) -> str:
        channels = self.get_channels()

        pci_channels = sim_base.Simulator.filter_channels_by_sys_type(
            channels, sys_pcie.PCIeChannel
        )
        pci_latency, pci_sync_period, pci_run_sync = (
            sim_base.Simulator.get_unique_latency_period_sync(pci_channels)
        )

        eth_channels = sim_base.Simulator.filter_channels_by_sys_type(channels, sys_eth.EthChannel)
        eth_latency, eth_sync_period, eth_run_sync = (
            sim_base.Simulator.get_unique_latency_period_sync(eth_channels)
        )

        if eth_run_sync != pci_run_sync:
            raise Exception(
                "currently using different synchronization values for pci and eth is not supported"
            )
        run_sync = eth_run_sync
        sync_period = min(pci_sync_period, eth_sync_period)

        cmd = f"{inst.join_repo_base(relative_path=self._executable)} "

        nic_devices = self.filter_components_by_type(ty=sys_nic.SimplePCIeNIC)
        assert len(nic_devices) == 1
        nic_device = nic_devices[0]

        socket = inst.update_get_socket(interface=nic_device._pci_if)
        assert socket is not None and socket._type == inst_socket.SockType.LISTEN
        cmd += f"{socket._path} "

        socket = inst.update_get_socket(interface=nic_device._eth_if)
        assert socket is not None and socket._type == inst_socket.SockType.LISTEN
        cmd += f"{socket._path} "

        cmd += (
            f" {inst.get_simulator_shm_pool_path(sim=self)} {int(run_sync)} {self._start_tick}"
            f" {sync_period} {pci_latency} {eth_latency}"
        )

        # if self.mac is not None:  # TODO: FIXME
        #     cmd += " " + ("".join(reversed(self.mac.split(":"))))

        if self.extra_args is not None:
            cmd += " " + self.extra_args

        return cmd


class I40eNicSim(NICSim):

    def __init__(self, simulation: sim_base.Simulation):
        super().__init__(
            simulation=simulation,
            executable="sims/nic/i40e_bm/i40e_bm",
        )
        self.name = f"NICSim-{self._id}"

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        return json_obj

    @classmethod
    def fromJSON(cls, simulation: sim_base.Simulation, json_obj: dict) -> I40eNicSim:
        return super().fromJSON(simulation, json_obj)

    def run_cmd(self, inst: inst_base.Instantiation) -> str:
        return super().run_cmd(inst=inst)


class CorundumBMNICSim(NICSim):
    def __init__(self, simulation: sim_base.Simulation):
        super().__init__(
            simulation=simulation,
            executable="sims/nic/corundum_bm/corundum_bm",
        )
        self.name = f"CorundumBMNICSim-{self._id}"

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        return json_obj

    @classmethod
    def fromJSON(cls, simulation: sim_base.Simulation, json_obj: dict) -> CorundumBMNICSim:
        return super().fromJSON(simulation, json_obj)

    def run_cmd(self, inst: inst_base.Instantiation) -> str:
        cmd = super().run_cmd(inst=inst)
        return cmd


class CorundumVerilatorNICSim(NICSim):

    def __init__(self, simulation: sim_base.Simulation):
        super().__init__(
            simulation=simulation,
            executable="sims/nic/corundum/corundum_verilator",
        )
        self.name = f"CorundumVerilatorNICSim-{self._id}"
        self.clock_freq = 250  # MHz

    def resreq_mem(self) -> int:
        # this is a guess
        return 512

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        json_obj["clock_freq"] = self.clock_freq
        return json_obj

    @classmethod
    def fromJSON(cls, simulation: sim_base.Simulation, json_obj: dict) -> CorundumVerilatorNICSim:
        return super().fromJSON(simulation, json_obj)

    def run_cmd(self, inst: inst_base.Instantiation) -> str:
        cmd = super().run_cmd(inst=inst)
        cmd += f" {self.clock_freq}"
        return cmd



# Copyright 2024 Max Planck Institute for Software Systems, and
# National University of Singapore
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


from simbricks.orchestration import system
from simbricks.utils import base as utils_base


def connect_host_and_device(
    host: system.Host, device: system.Component
) -> system.pcie.PCIeChannel:
    utils_base.has_expected_type(obj=host, expected_type=system.Host)
    utils_base.has_expected_type(obj=device, expected_type=system.Component)

    host_interface = system.pcie.PCIeHostInterface(c=host)
    host.add_if(interface=host_interface)

    device_interface = system.pcie.PCIeDeviceInterface(c=device)
    device.add_if(interface=device_interface)

    pcie_channel = system.pcie.PCIeChannel(host=host_interface, dev=device_interface)
    return pcie_channel


def connect_eth_devices(
    device_a: system.Component, device_b: system.Component
) -> system.EthChannel:
    utils_base.has_expected_type(obj=device_a, expected_type=system.Component)
    utils_base.has_expected_type(obj=device_b, expected_type=system.Component)

    eth_inter_a = system.eth.EthInterface(c=device_a)
    device_a.add_if(interface=eth_inter_a)

    eth_inter_b = system.eth.EthInterface(c=device_b)
    device_b.add_if(interface=eth_inter_b)

    eth_channel = system.eth.EthChannel(a=eth_inter_a, b=eth_inter_b)
    return eth_channel


def install_app(
    host: system.Host, app_ty: system.Application, **kwargs
) -> system.Application:
    utils_base.has_expected_type(obj=host, expected_type=system.Host)
    utils_base.has_expected_type(obj=app_ty, expected_type=system.Application)

    application = app_ty(h=host, **kwargs)
    host.add_app(a=application)

    return application



# Copyright 2024 Max Planck Institute for Software Systems, and
# National University of Singapore
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import annotations

import abc
import asyncio
import typing as tp
import simbricks.orchestration.system as sys_conf
import simbricks.orchestration.instantiation.base as inst_base
import simbricks.orchestration.instantiation.socket as inst_socket
import simbricks.orchestration.simulation.channel as sim_chan
import simbricks.utils.base as utils_base

if tp.TYPE_CHECKING:
    from simbricks.orchestration.simulation import (
        Channel,
        base as sim_base,
    )


class Simulator(utils_base.IdObj, abc.ABC):
    """Base class for all simulators."""

    def __init__(
        self,
        simulation: sim_base.Simulation,
        executable: str,
        name: str = "",
    ) -> None:
        super().__init__()
        self.name: str = name
        self._executable: str = executable
        self._simulation: sim_base.Simulation = simulation
        self._components: set[sys_conf.Component] = set()
        self._wait: bool = False
        self._start_tick: int = 0
        """The timestamp at which to start the simulation. This is useful when
        the simulator is only attached at a later point in time and needs to
        synchronize with connected simulators. For example, this could be used
        when taking checkpoints to only attach certain simulators after the
        checkpoint has been taken."""
        self._extra_args: str | None = None
        simulation.add_sim(self)

    T = tp.TypeVar("T")

    def filter_components_by_pred(
        self,
        pred: tp.Callable[[sys_conf.Component], bool],
        ty: type[T] = sys_conf.Component,
    ) -> list[T]:
        return list(filter(pred, self._components))

    def filter_components_by_type(self, ty: type[T]) -> list[T]:
        return self.filter_components_by_pred(pred=lambda comp: isinstance(comp, ty), ty=ty)

    @property
    def extra_args(self) -> str:
        return self._extra_args

    @extra_args.setter
    def extra_args(self, extra_args: str):
        self._extra_args = extra_args

    @property
    def wait_terminate(self) -> bool:
        if self._wait:
            return True
        host_comps = self.filter_components_by_type(ty=sys_conf.Host)
        for host in host_comps:
            for app in host.applications:
                if app.wait:
                    return True
        return False

    @wait_terminate.setter
    def wait_terminate(self, wait: bool):
        self._wait = wait

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        json_obj["name"] = self.name
        json_obj["executable"] = self._executable
        json_obj["simulation"] = self._simulation.id()

        components_json = []
        for comp in self._components:
            components_json.append(comp.id())
        json_obj["components"] = components_json

        json_obj["wait"] = self._wait
        json_obj["start_tick"] = self._start_tick
        json_obj["extra_args"] = self._extra_args
        return json_obj

    @classmethod
    def fromJSON(cls, simulation: Simulation, json_obj: dict) -> Simulator:
        instance = super().fromJSON(json_obj)
        instance.name = utils_base.get_json_attr_top(json_obj, "name")
        instance._executable = utils_base.get_json_attr_top(json_obj, "executable")

        sim_id = int(utils_base.get_json_attr_top(json_obj, "simulation"))
        assert sim_id == simulation.id()
        instance._simulation = simulation

        instance._components = set()
        components_json = utils_base.get_json_attr_top(json_obj, "components")
        for comp_id in components_json:
            component = simulation.system.get_comp(comp_id)
            Simulator.add(instance, component)
            # instance.add(component)

        instance._wait = bool(utils_base.get_json_attr_top(json_obj, "wait"))
        instance._start_tick = int(utils_base.get_json_attr_top(json_obj, "start_tick"))
        instance._extra_args = utils_base.get_json_attr_top_or_none(json_obj, "extra_args")

        return instance

    @staticmethod
    def filter_sockets(
        sockets: list[inst_socket.Socket],
        filter_type: inst_socket.SockType = inst_socket.SockType.LISTEN,
    ) -> list[inst_socket.Socket]:
        res = list(filter(lambda sock: sock._type == filter_type, sockets))
        return res

    @staticmethod
    def split_sockets_by_type(
        sockets: list[inst_socket.Socket],
    ) -> tuple[list[inst_socket.Socket], list[inst_socket.Socket]]:
        listen = Simulator.filter_sockets(sockets=sockets, filter_type=inst_socket.SockType.LISTEN)
        connect = Simulator.filter_sockets(
            sockets=sockets, filter_type=inst_socket.SockType.CONNECT
        )
        return listen, connect

    # helper method for simulators that do not support
    # multiple sync periods etc. Should become eventually
    # at some point in the future...
    @staticmethod
    def get_unique_latency_period_sync(
        channels: list[sim_chan.Channel],
    ) -> tuple[int, int, bool]:
        latency = None
        sync_period = None
        run_sync = False
        for channel in channels:
            sync_period = (
                min(sync_period, channel.sync_period) if sync_period else channel.sync_period
            )
            run_sync = run_sync or channel._synchronized
            latency = (
                max(latency, channel.sys_channel.latency)
                if latency
                else channel.sys_channel.latency
            )
        if latency is None or sync_period is None:
            raise Exception("could not determine eth_latency and sync_period")
        return latency, sync_period, run_sync

    def components(self) -> set[sys_conf.Component]:
        return self._components

    def resreq_cores(self) -> int:
        """
        Number of cores this simulator requires during execution.

        This is used for scheduling multiple runs and experiments.
        """
        return 1

    def resreq_mem(self) -> int:
        """
        Number of memory in MB this simulator requires during execution.

        This is used for scheduling multiple runs and experiments.
        """
        return 64

    def full_name(self) -> str:
        """Full name of the simulator."""
        return ""

    def add(self, comp: sys_conf.Component) -> None:
        if comp in self._components:
            raise Exception("cannot add the same specification twice to a simulator")
        self._components.add(comp)
        self._simulation.add_spec_sim_map(comp, self)

    def _chan_needs_instance(self, chan: sys_conf.Channel) -> bool:
        if chan.a.component in self._components and chan.b.component in self._components:
            return False
        return True

    def _get_socks_by_comp(
        self, inst: inst_base.Instantiation, comp: sys_conf.Component
    ) -> list[inst_socket.Socket]:
        if comp not in self._components:
            raise Exception("comp must be a simulators component")
        sockets = []
        for interface in comp.interfaces():
            socket = inst.update_get_socket(interface=interface)
            if socket:
                sockets.append(socket)
        return sockets

    def _get_socks_by_all_comp(self, inst: inst_base.Instantiation) -> list[inst_socket.Socket]:
        sockets = []
        for comp in self._components:
            sockets.extend(self._get_socks_by_comp(inst=inst, comp=comp))
        return sockets

    def _get_all_sockets_by_type(
        self, inst: inst_base.Instantiation, sock_type: inst_socket.SockType
    ) -> list[inst_socket.Socket]:
        sockets = self._get_socks_by_all_comp(inst=inst)
        sockets = Simulator.filter_sockets(sockets=sockets, filter_type=sock_type)
        return sockets

    def _get_channel(self, chan: sys_conf.Channel) -> sim_chan.Channel | None:
        if self._chan_needs_instance(chan):
            return self._simulation.retrieve_or_create_channel(chan=chan)
        return None

    def get_channels(self) -> list[sim_chan.Channel]:
        channels = []
        for comp_spec in self._components:
            comp_sys_channels = comp_spec.channels()
            for chan in comp_sys_channels:
                channel = self._get_channel(chan=chan)
                if channel is None:
                    continue
                channels.append(channel)
        return channels

    @staticmethod
    def filter_channels_by_sys_type(channels: list[sim_chan.Channel], ty: type[T]) -> list[T]:
        filtered = list(filter(lambda chan: isinstance(chan.sys_channel, ty), channels))
        return filtered

    # pylint: disable=unused-argument
    @abc.abstractmethod
    def run_cmd(self, inst: inst_base.Instantiation) -> str:
        """Command to execute this simulator."""
        raise Exception("must be implemented in sub-class")

    def checkpoint_commands(self) -> list[str]:
        return []

    def cleanup_commands(self) -> list[str]:
        return []

    @abc.abstractmethod
    def supported_socket_types(self, interface: sys_conf.Interface) -> set[inst_socket.SockType]:
        return set()

    # Sockets to be cleaned up: always the CONNECTING sockets
    # pylint: disable=unused-argument
    def sockets_cleanup(self, inst: inst_base.Instantiation) -> list[inst_socket.Socket]:
        return self._get_all_sockets_by_type(inst=inst, sock_type=inst_socket.SockType.LISTEN)

    # sockets to wait for indicating the simulator is ready
    # pylint: disable=unused-argument
    def sockets_wait(self, inst: inst_base.Instantiation) -> list[inst_socket.Socket]:
        return self._get_all_sockets_by_type(inst=inst, sock_type=inst_socket.SockType.LISTEN)

    def start_delay(self) -> int:
        return 5

    def supports_checkpointing(self) -> bool:
        return False

    async def prepare(self, inst: inst_base.Instantiation) -> None:
        promises = [comp.prepare(inst=inst) for comp in self._components]
        await asyncio.gather(*promises)


class Simulation(utils_base.IdObj):
    """
    Base class for all simulation experiments.

    Contains the simulators to be run and experiment-wide parameters.
    """

    def __init__(self, name: str, system: sys_conf.System) -> None:
        super().__init__()
        self.name = name
        """
        This experiment's name.

        Can be used to run only a selection of experiments.
        """
        self.system: sys_conf.System = system
        self.timeout: int | None = None
        """Timeout for experiment in seconds."""
        self.metadata: dict[str, tp.Any] = {}

        self._sys_sim_map: dict[sys_conf.Component, Simulator] = {}
        """System component and its simulator pairs"""
        self._sim_list: list[Simulator] = []
        """Channel spec and its instanciation"""

        self._chan_map: dict[sys_conf.Channel, sim_chan.Channel] = {}
        """Channel spec and its instanciation"""

    def toJSON(self) -> dict:
        json_obj = super().toJSON()

        json_obj["name"] = self.name
        json_obj["metadata"] = self.metadata
        json_obj["system"] = self.system.id()
        json_obj["timeout"] = self.timeout

        simulators_json = []
        for sim in self._sim_list:
            utils_base.has_attribute(sim, "toJSON")
            simulators_json.append(sim.toJSON())

        json_obj["sim_list"] = simulators_json

        # TODO: we do not need this --> we can create it implicitly when deserializing because the simulators store a list of components themselves
        # sys_sim_map_json = []
        # for comp, sim in self._sys_sim_map.items():
        #     sys_sim_map_json.append([comp.id(), sim.id()])

        # json_obj["sys_sim_map"] = sys_sim_map_json

        chan_map_json = []
        chan_json = []
        for (
            sys_chan,
            sim_chan,
        ) in self._chan_map.items():
            utils_base.has_attribute(sim_chan, "toJSON")
            chan_json = sim_chan.toJSON()
            chan_map_json.append([sys_chan.id(), chan_json])
            # chan_json.append(sim_chan.toJSON())

        json_obj["chan_map"] = chan_map_json
        # json_obj["simulation_channels"] = chan_json

        return json_obj

    @classmethod
    def fromJSON(cls, system: sys_conf.System, json_obj: dict) -> Simulation:
        instance = super().fromJSON(json_obj)
        instance.metadata = utils_base.get_json_attr_top(json_obj, "metadata")
        instance.name = utils_base.get_json_attr_top(json_obj, "name")
        system_id = int(utils_base.get_json_attr_top(json_obj, "system"))
        assert system_id == system.id()
        instance.system = system
        instance.timeout = utils_base.get_json_attr_top_or_none(json_obj, "timeout")

        instance._sim_list = []
        instance._sys_sim_map = {}
        simulators_json = utils_base.get_json_attr_top(json_obj, "sim_list")
        for sim_json in simulators_json:
            sim_class = utils_base.get_cls_by_json(sim_json)
            utils_base.has_attribute(sim_class, "fromJSON")
            sim = sim_class.fromJSON(instance, sim_json)
            instance._sim_list.append(sim)

        # TODO: probably do not need this --> we create it when deserializing the simulators that store a list of components themselves
        # instance._sys_sim_map: dict[sys_conf.Component, Simulator] = {}
        # sys_sim_map_json = utils_base.get_json_attr_top(json_obj, "sys_sim_map")
        # print(sys_sim_map_json)
        # for sys_id, sim_id in sys_sim_map_json:
        #     print(f"{sys_id} --> {sim_id}")
        #     # TODO
        #     pass

        instance._chan_map = {}
        chan_map_json = utils_base.get_json_attr_top(json_obj, "chan_map")
        for sys_id, chan_json in chan_map_json:
            chan_class = utils_base.get_cls_by_json(chan_json)
            utils_base.has_attribute(chan_class, "fromJSON")
            sim_chan = chan_class.fromJSON(instance, chan_json)
            sys_chan = instance.system.get_chan(sys_id)
            instance.update_channel_mapping(sys_chan=sys_chan, sim_chan=sim_chan)

        return instance

    def add_sim(self, sim: Simulator):
        if sim in self._sim_list:
            raise Exception("Simulaotr is already added")
        self._sim_list.append(sim)

    def add_spec_sim_map(self, sys: sys_conf.Component, sim: Simulator):
        """Add a mapping from specification to simulation instance"""
        if sys in self._sys_sim_map:
            raise Exception("system component is already mapped by simulator")
        self._sys_sim_map[sys] = sim

    def is_channel_instantiated(self, chan: sys_conf.Channel) -> bool:
        return chan in self._chan_map

    def update_channel_mapping(
        self, sys_chan: sys_conf.Channel, sim_chan: sim_chan.Channel
    ) -> None:
        if self.is_channel_instantiated(sys_chan):
            raise Exception(
                f"channel {sys_chan} is already mapped. Cannot insert mapping {sys_chan.id()} -> {sim_chan.id()}"
            )
        self._chan_map[sys_chan] = sim_chan

    def retrieve_or_create_channel(self, chan: sys_conf.Channel) -> sim_chan.Channel:
        if self.is_channel_instantiated(chan):
            return self._chan_map[chan]

        channel = sim_chan.Channel(chan)
        self.update_channel_mapping(sys_chan=chan, sim_chan=channel)
        return channel

    def get_channel(self, chan: sys_conf.Channel) -> sim_chan.Channel:
        if not self.is_channel_instantiated(chan):
            raise RuntimeError(f"Channel {chan} is not instantiated")
        return self._chan_map[chan]

    def all_simulators(self) -> list[Simulator]:
        return self._sim_list

    def get_all_channels(self, lazy: bool = False) -> list[Channel]:
        if lazy:
            return list(self._chan_map.values())

        all_channels = []
        for sim in self.all_simulators():
            channels = sim.get_channels()
            all_channels.extend(channels)
        return all_channels

    def enable_synchronization(
        self, amount: int | None = None, ratio: utils_base.Time | None = None
    ) -> None:
        for chan in self.get_all_channels():
            chan._synchronized = True
            if amount and ratio:
                chan.set_sync_period(amount=amount, ratio=ratio)

    def resreq_mem(self) -> int:
        """Memory required to run all simulators in this experiment."""
        mem = 0
        for s in self.all_simulators():
            mem += s.resreq_mem()
        return mem

    def resreq_cores(self) -> int:
        """Number of Cores required to run all simulators in this experiment."""
        cores = 0
        for s in self.all_simulators():
            cores += s.resreq_cores()
        return cores

    def find_sim(self, comp: sys_conf.Component) -> sim_base.Simulator:
        """Returns the used simulator object for the system component."""
        utils_base.has_expected_type(comp, sys_conf.Component)
        if comp not in self._sys_sim_map:
            raise Exception(f"Simulator not found for component: {comp}")
        return self._sys_sim_map[comp]

    async def prepare(self, inst: inst_base.Instantiation) -> None:
        promises = []
        for sim in inst.fragment.all_simulators():
            promises.append(sim.prepare(inst=inst))
        await asyncio.gather(*promises)

    def any_supports_checkpointing(self) -> bool:
        if len(list(filter(lambda sim: sim.supports_checkpointing(), self._sim_list))) > 0:
            return True
        return False
